#!/usr/bin/env python

import nibabel as nib
import nilearn.image
import numpy as np
import argparse
import os
from qsmxt.scripts.qsmxt_functions import extend_fname
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, traits, InputMultiPath

def resample_to_reference(in_file, ref_file, out_file=None, interpolation='continuous'):
    print(f"Resampling {in_file} to {ref_file} with interpolation {interpolation}")

    # Load NIfTI files
    in_nii = nib.load(in_file)
    ref_nii = nib.load(ref_file)

    print("Loaded input and reference files")

    # Check if the input image is already aligned with the reference image
    if np.array_equal(in_nii.affine, ref_nii.affine):
        print("Input image is already aligned with the reference image")
        return in_file

    # Resample the source image to match the reference image
    resampled_nii = nilearn.image.resample_img(
        in_nii, 
        target_affine=ref_nii.affine, 
        target_shape=np.array(ref_nii.header.get_data_shape()),
        interpolation=interpolation
    )

    print("Resampled the image")

    # Save the resampled image
    out_file = out_file or extend_fname(in_file, "_resampled", out_dir=os.getcwd())
    nib.save(resampled_nii, out_file)
    print(f"Saved resampled image to {out_file}")

    return out_file

class ResampleLikeInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True)
    ref_file = InputMultiPath(File, mandatory=True, exists=True)
    interpolation = traits.Enum("continuous", "nearest", usedefault=True, default="continuous")

class ResampleLikeOutputSpec(TraitedSpec):
    out_file = File()

class ResampleLikeInterface(SimpleInterface):
    input_spec = ResampleLikeInputSpec
    output_spec = ResampleLikeOutputSpec

    def _run_interface(self, runtime):
        if isinstance(self.inputs.ref_file, list):
            ref_file = self.inputs.ref_file[0]
        else:
            ref_file = self.inputs.ref_file

        out_file = resample_to_reference(
            in_file=self.inputs.in_file,
            ref_file=ref_file,
            interpolation=self.inputs.interpolation
        )

        if out_file is not None:
            self._results['out_file'] = out_file
        else:
            raise RuntimeError("Resampling failed, out_file is None")

        return runtime



if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Resample a NIfTI image to match the dimensions, resolution, and orientation of a reference image.")
    parser.add_argument("in_file", help="Path to the source NIfTI file to be resampled.")
    parser.add_argument("ref_file", help="Path to the reference NIfTI file.")
    parser.add_argument("out_file", help="Path to save the resampled NIfTI file.")
    parser.add_argument("--interpolation", choices=['nearest', 'continuous'], default='continuous', 
                        help="Type of interpolation for resampling ('nearest' or 'continuous'). Default is 'continuous'.")
    args = parser.parse_args()

    resample_to_reference(args.in_file, args.ref_file, args.out_file, args.interpolation)




import os

from nipype.interfaces.base import TraitedSpec, File, traits
from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from qsmxt.scripts import qsmxt_functions


class LaplacianUnwrappingInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(LaplacianUnwrappingInputSpec, self).__init__(**inputs)
    phase = File(
        exists=True,
        mandatory=True,
        argstr="--phase %s",
        position=0
    )
    mask = File(
        exists=True,
        mandatory=True,
        argstr="--mask %s",
        position=1
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    phase_unwrapped = File(
        argstr="--unwrapped-phase-out %s",
        name_source=['phase'],
        name_template='%s_unwrapped-laplacian.nii',
        position=3
    )


class LaplacianUnwrappingOutputSpec(TraitedSpec):
    phase_unwrapped = File(exists=True)


class LaplacianUnwrappingInterface(CommandLineJulia):
    def __init__(self, **inputs): super(LaplacianUnwrappingInterface, self).__init__(**inputs)
    input_spec = LaplacianUnwrappingInputSpec
    output_spec = LaplacianUnwrappingOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_laplacian_unwrapping.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["phase_unwrapped"] = qsmxt_functions.extend_fname(self.inputs.phase, "_unwrapped-laplacian", ext="nii", out_dir=os.getcwd())
        return outputs


class PhaseToFreqInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(PhaseToFreqInputSpec, self).__init__(**inputs)
    phase = File(
        exists=True,
        mandatory=True,
        argstr="--phase %s",
        position=0
    )
    TE = traits.Float(
        argstr="--TEs [%s]",
        position=2
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    B0 = traits.Float(
        argstr="--b0-str %s",
        default=3,
        position=4
    )
    frequency = File(
        argstr="--frequency-out %s",
        name_source=['phase'],
        name_template='%s_freq.nii',
        position=5
    )


class PhaseToFreqOutputSpec(TraitedSpec):
    frequency = File(exists=True)


class PhaseToFreqInterface(CommandLineJulia):
    def __init__(self, **inputs): super(PhaseToFreqInterface, self).__init__(**inputs)
    input_spec = PhaseToFreqInputSpec
    output_spec = PhaseToFreqOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_phase_to_frequency.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["frequency"] = qsmxt_functions.extend_fname(self.inputs.frequency, "_freq", ext="nii", out_dir=os.getcwd())
        return outputs


class VsharpInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(VsharpInputSpec, self).__init__(**inputs)
    frequency = File(
        exists=True,
        mandatory=True,
        argstr="--frequency %s",
        position=0
    )
    mask = File(
        exists=True,
        mandatory=True,
        argstr="--mask %s",
        position=1
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    tissue_frequency = File(
        argstr="--tissue-frequency-out %s",
        name_source=['frequency'],
        name_template='%s_vsharp.nii',
        position=3
    )
    vsharp_mask = File(
        argstr="--vsharp-mask-out %s",
        name_source=['mask'],
        name_template='%s_vsharp-mask.nii',
        position=4
    )


class VsharpOutputSpec(TraitedSpec):
    tissue_frequency = File(exists=True)
    vsharp_mask = File(exists=True)


class VsharpInterface(CommandLineJulia):
    def __init__(self, **inputs): super(VsharpInterface, self).__init__(**inputs)
    input_spec = VsharpInputSpec
    output_spec = VsharpOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_vsharp.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["tissue_frequency"] = qsmxt_functions.extend_fname(self.inputs.frequency, "_vsharp", ext="nii", out_dir=os.getcwd())
        outputs["vsharp_mask"] = qsmxt_functions.extend_fname(self.inputs.mask, "_vsharp-mask", ext="nii", out_dir=os.getcwd())
        return outputs


class PdfInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(PdfInputSpec, self).__init__(**inputs)
    frequency = File(
        exists=True,
        mandatory=True,
        argstr="--frequency %s",
        position=0
    )
    mask = File(
        exists=True,
        mandatory=True,
        argstr="--mask %s",
        position=1
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    tissue_frequency = File(
        argstr="--tissue-frequency-out %s",
        name_source=['frequency'],
        name_template='%s_pdf.nii',
        position=3
    )


class PdfOutputSpec(TraitedSpec):
    tissue_frequency = File(exists=True)


class PdfInterface(CommandLineJulia):
    def __init__(self, **inputs): super(PdfInterface, self).__init__(**inputs)
    input_spec = PdfInputSpec
    output_spec = PdfOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_pdf.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["tissue_frequency"] = qsmxt_functions.extend_fname(self.inputs.frequency, "_pdf", ext="nii", out_dir=os.getcwd())
        return outputs


class RtsQsmInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(RtsQsmInputSpec, self).__init__(**inputs)
    tissue_frequency = File(
        exists=True,
        mandatory=True,
        argstr="--tissue-frequency %s",
        position=0
    )
    mask = File(
        exists=True,
        mandatory=True,
        argstr="--mask %s",
        position=1
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    b0_direction = traits.ListFloat(
        argstr="--b0-dir '[%s]'",
        default=[0,0,1],
        position=3
    )
    qsm = File(
        argstr="--qsm-out %s",
        name_source=['tissue_frequency'],
        name_template='%s_rts.nii',
        position=4
    )


class RtsQsmOutputSpec(TraitedSpec):
    qsm = File(exists=True)


class RtsQsmInterface(CommandLineJulia):
    def __init__(self, **inputs): super(RtsQsmInterface, self).__init__(**inputs)
    input_spec = RtsQsmInputSpec
    output_spec = RtsQsmOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_rts.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["qsm"] = qsmxt_functions.extend_fname(self.inputs.tissue_frequency, "_rts", ext="nii", out_dir=os.getcwd())
        return outputs

class TvQsmInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(TvQsmInputSpec, self).__init__(**inputs)
    tissue_frequency = File(
        exists=True,
        mandatory=True,
        argstr="--tissue-frequency %s",
        position=0
    )
    mask = File(
        exists=True,
        mandatory=True,
        argstr="--mask %s",
        position=1
    )
    vsz = traits.ListFloat(
        argstr="--vsz '[%s]'",
        default=[1, 1, 1],
        position=2
    )
    b0_direction = traits.ListFloat(
        argstr="--b0-dir '[%s]'",
        default=[0,0,1],
        position=3
    )
    qsm = File(
        argstr="--qsm-out %s",
        name_source=['tissue_frequency'],
        name_template='%s_tv.nii',
        position=4
    )


class TvQsmOutputSpec(TraitedSpec):
    qsm = File(exists=True)


class TvQsmInterface(CommandLineJulia):
    def __init__(self, **inputs): super(TvQsmInterface, self).__init__(**inputs)
    input_spec = TvQsmInputSpec
    output_spec = TvQsmOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "qsmjl_tv.jl")

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["qsm"] = qsmxt_functions.extend_fname(self.inputs.tissue_frequency, "_tv", ext="nii", out_dir=os.getcwd())
        return outputs




#!/usr/bin/env python3

import argparse
import os
import sys
import json
import shutil
import datetime
import curses
import re
import pandas as pd

from dicompare import load_nifti_session

from qsmxt.scripts.qsmxt_functions import get_qsmxt_version, get_diff
from qsmxt.scripts.logger import LogLevel, make_logger, show_warning_summary

def script_exit(exit_code=0):
    logger = make_logger()
    show_warning_summary(logger)
    logger.log(LogLevel.INFO.value, 'Finished')
    exit(exit_code)

def interactive_table(session: pd.DataFrame):
    """
    A curses-based interface with:
      - A "regex row" at row_idx = -1, with one regex input per column (except NIfTI_Path).
      - A data table (row_idx >= 0) with columns:
         0) NIfTI_Path (read-only, toggles full/basename on TAB)
         1) sub
         2) ses
         3) acq
         4) run
         5) echo
         6) part
         7) suffix
         8) MagneticFieldStrength
         9) EchoTime

    Controls:
      - ARROWS = move, ENTER = apply regex (if on regex row) or move down (if on data table)
      - ESC = finish
      - TAB = toggle between showing full path or basename in NIfTI_Path column
      - BACKSPACE = remove last character from the current field (if not empty),
        or if already empty, copy from the field above (if available).
      - DELETE = clear (empty) the current field if not empty, or if already empty, copy from above.
      - We also show context-specific help for each column in a line above the table.
    """

    session.reset_index(drop=True, inplace=True)

    columns = [
        "NIfTI_Path",
        "sub",
        "ses",
        "acq",
        "run",
        "echo",
        "part",
        "suffix",
        "MagneticFieldStrength",
        "EchoTime",
    ]

    # For toggling whether we display full path or basename
    show_full_path = True
    row_idx = -1  # -1 => "regex row"; 0.. => data table
    col_idx = 0

    # Ensure all columns exist
    for c in columns:
        if c not in session.columns:
            session[c] = None

    # Turn all non-None values to strings
    for c in columns:
        session[c] = session[c].apply(lambda x: str(x) if pd.notnull(x) else None)

    # Keep a hidden full path for applying regex
    session["NIfTI_Path_Full"] = session["NIfTI_Path"]

    # Regex patterns for each column (except NIfTI_Path)
    regex_map = {}
    for c in columns:
        if c != "NIfTI_Path":
            regex_map[c] = ""

    # Default guesses
    default_patterns = {
        "sub": r"sub-([A-Za-z0-9]+)",
        "ses": r"ses-([A-Za-z0-9]+)",
        "acq": r"acq-([A-Za-z0-9]+)",
        "run": r"run-([0-9]+)",
        "echo": r"echo-([0-9]+)",
        "part": r"part-([A-Za-z]+)",
        "suffix": r"_([A-Za-z0-9]+)\.nii",
        "MagneticFieldStrength": r"_B(\d+\.\d+)",
        "EchoTime": r"_TE(\d+\.\d+)",
    }
    for k,v in default_patterns.items():
        if k in regex_map:
            regex_map[k] = v

    column_help = {
        "NIfTI_Path": "[READ-ONLY]: Press TAB to toggle full path or basename.",
        "sub": "[REQUIRED]: Subject ID e.g. 1 (BIDS: sub-<label>).",
        "ses": "[OPTIONAL]: Session date e.g. 20241231 (BIDS: ses-<label>).",
        "acq": "[OPTIONAL]: Acquisition label (BIDS: acq-<label>).",
        "run": "[REQUIRED for multiple runs]: For multiple runs of an acquisition in a single session e.g. 1, 2, etc.",
        "echo": "[REQUIRED for MEGRE]: Counter for identifying echo chronology.",
        "part": "[REQUIRED for T2starw or MEGRE]: Identifies signal part ('mag', 'phase', 'real', or 'imag').",
        "suffix": "[REQUIRED]: Identifies acquisition type T2starw or MEGRE for single or multi-echo GRE, respectively, T1w for anatomical T1.",
        "MagneticFieldStrength": "[REQUIRED for T2starw/MEGRE]: Magnetic field strength in Teslas e.g. 1.5, 3, 7.",
        "EchoTime": "[REQUIRED for T2starw/MEGRE]: Echo time in seconds e.g. 0.015, 0.03."
    }

    def apply_regex_to_filenames(column):
        pattern = regex_map[column]
        if not pattern:
            return False, f"No regex pattern for {column}."

        try:
            compiled = re.compile(pattern)
        except re.error as e:
            return False, f"Invalid regex for {column}: {e}"

        assigned_count = 0
        for i in range(len(session)):
            fullp = str(session.loc[i, "NIfTI_Path_Full"])
            match = compiled.search(fullp)
            if match:
                val = match.group(1)
                session.loc[i, column] = val
                assigned_count += 1

        return True, f"Regex assigned {column} for {assigned_count} file(s)."

    def copy_from_above_if_exists(r, c):
        """
        Copies the value from row (r-1, c) to (r, c) if (r > 0) and
        the above cell is not empty.
        """
        if r <= 0:
            return  # can't copy from above if we're in the top row or negative
        above_val = session.loc[r-1, c]
        if pd.notnull(above_val) and str(above_val).strip() != "":
            session.loc[r, c] = above_val

    def validate_data(data):
        errors = []
        # col names
        col_names = ["NIfTI_Path", "sub", "ses", "acq", "run", "echo", "part", "suffix", "MagneticFieldStrength", "EchoTime"]
        for i, row in data.iterrows():
            row = row[col_names].drop("NIfTI_Path")
            if row.isnull().all():
                continue

            for c in ["sub", "suffix"]:
                if pd.isnull(row[c]):
                    errors.append((i, col_names.index(c), "Subject ID and suffix are required."))
            
            # if suffix is T2starw or MEGRE, MagneticFieldStrength and EchoTime are required
            if row["suffix"] in ["T2starw", "MEGRE"]:
                if pd.isnull(row["MagneticFieldStrength"]):
                    errors.append((i, col_names.index("MagneticFieldStrength"), "T2starw and MEGRE images require MagneticFieldStrength."))
                if pd.isnull(row["EchoTime"]):
                    errors.append((i, col_names.index("EchoTime"), "T2starw and MEGRE images require EchoTime."))
                
            # if suffix is MEGRE, echo is required
            if row["suffix"] == "MEGRE":
                if pd.isnull(row["echo"]):
                    errors.append((i, col_names.index("echo"), "MEGRE images require echo."))
            
            # if suffix is T2starw or MEGRE, part is required
            if row["suffix"] in ["T2starw", "MEGRE"]:
                if pd.isnull(row["part"]):
                    errors.append((i, col_names.index("part"), "T2starw and MEGRE images require part."))
            
            # if suffix is T2starw, echo is not allowed
            if row["suffix"] == "T2starw":
                if pd.notnull(row["echo"]):
                    errors.append((i, col_names.index("echo"), "T2starw images should not have echo."))

            # part values may only be 'mag', 'phase', 'real', or 'imag'
            if pd.notnull(row["part"]) and str(row["part"]).strip() not in ["mag", "phase", "real", "imag"]:
                errors.append((i, col_names.index("part"), "part should be 'mag', 'phase', 'real', or 'imag'."))

            # suffix values may only be 'T2starw', 'MEGRE', or 'T1w'
            if pd.notnull(row["suffix"]) and str(row["suffix"]).strip() not in ["T2starw", "MEGRE", "T1w"]:
                errors.append((i, col_names.index("suffix"), "suffix should be 'T2starw', 'MEGRE', or 'T1w'."))
            
            # echo values should be positive integers
            if pd.notnull(row["echo"]):
                try:
                    val = int(row["echo"])
                    if val <= 0:
                        errors.append((i, col_names.index("echo"), "echo should be a positive integer."))
                except ValueError:
                    errors.append((i, col_names.index("echo"), "echo should be a positive integer."))

            # MagneticFieldStrength and EchoTime should be positive floats
            for c in ["MagneticFieldStrength", "EchoTime"]:
                if pd.notnull(row[c]):
                    try:
                        val = float(row[c])
                        if val <= 0:
                            errors.append((i, col_names.index(c), f"{c} should be a positive float."))
                    except ValueError:
                        errors.append((i, col_names.index(c), f"{c} should be a positive float."))

        # the same combination of sub, ses, acq, run, echo, part, and suffix should not be repeated
        dupes = data.duplicated(subset=["sub", "ses", "acq", "run", "echo", "part", "suffix"], keep=False)
        for i, dupe in dupes.items():
            if dupe:
                errors.append((i, col_names.index("sub"), "Duplicate combination of sub, ses, acq, run, echo, part, and suffix."))

        # 2) For a given sub/ses/acq/run combo, MagneticFieldStrength should be constant
        group_mf = data.groupby(["sub","ses","acq","run"], dropna=False)
        for group_keys, grp in group_mf:
            # sub, ses, acq, run => all rows must share the same MagneticFieldStrength (if not null)
            non_null_values = grp["MagneticFieldStrength"].dropna().unique()
            # If there's more than 1 distinct non-null MF strength => error on each row
            if len(non_null_values) > 1:
                for idx in grp.index:
                    errors.append((idx, col_names.index("MagneticFieldStrength"),
                                "MagneticFieldStrength must be constant for sub/ses/acq/run."))

        # 3) For a given sub/ses/acq/run/echo combo, EchoTime should be constant
        group_et = data.groupby(["sub","ses","acq","run","echo"], dropna=False)
        for group_keys, grp in group_et:
            # sub, ses, acq, run, echo => all rows must share the same EchoTime (if not null)
            non_null_times = grp["EchoTime"].dropna().unique()
            if len(non_null_times) > 1:
                for idx in grp.index:
                    errors.append((idx, col_names.index("EchoTime"),
                                "EchoTime must be constant for sub/ses/acq/run/echo."))
        
        # --- 3) If multiple rows share the same sub/ses/acq/run/part/suffix,
        #         their echo values must be consecutive 1..n.
        #         (Ignoring T2starw that disallows echo anyway).
        grouping_cols = ["sub","ses","acq","run","part","suffix"]
        grouped = data.groupby(grouping_cols, dropna=False)  # dropna=False so we keep invalid combos too
        for group_keys, grp in grouped:
            suffix_val = group_keys[-1]  # suffix is last in group_keys
            if suffix_val != "MEGRE":
                continue

            echo_vals = []
            row_indices = []
            for idx, thisrow in grp.iterrows():
                e = str(thisrow["echo"]).strip() if pd.notnull(thisrow["echo"]) else ""
                # if e is not empty
                if e:
                    try:
                        echo_vals.append(int(e))
                    except ValueError:
                        # already flagged above
                        pass
                row_indices.append(idx)

            sorted_e = sorted(echo_vals)
            if sorted_e != list(range(1, len(sorted_e) + 1)):
                for idx in row_indices:
                    errors.append((idx, col_names.index("echo"), "For multiple echoes, they must start at 1 and be consecutive within the same sub/ses/acq/run/part/suffix."))

        if errors:
            return False, errors
        return True, []

    def curses_ui(stdscr):
        nonlocal row_idx, col_idx, show_full_path

        curses.curs_set(1)
        curses.start_color()

        # normal
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
        normal = curses.color_pair(1)

        # error
        curses.init_pair(2, curses.COLOR_RED, curses.COLOR_BLACK)
        error = curses.color_pair(2)

        # warning
        curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        warning = curses.color_pair(3)

        # success
        curses.init_pair(4, curses.COLOR_GREEN, curses.COLOR_BLACK)
        success = curses.color_pair(4)

        valid, errors = validate_data(session)
        if not valid:
            row_idx, col_idx, message = errors[0]
            status = (error, message)
        else:
            status = (success, "Valid")

        nrows = len(session)
        ncols = len(columns)
        changes = False

        while True:
            stdscr.clear()
            max_y, max_x = stdscr.getmaxyx()

            # Layout plan:
            #  Line 0-1 : instructions
            #  Line 2   : status
            #  Line 3   : help text for current column
            #  Line 4   : BLANK
            #  Line 5   : regex header
            #  Line 6   : regex row
            #  Line 7   : BLANK
            #  Line 8   : table header
            #  Line 9.. : data rows

            # Quick naive sizing for columns
            col0_width = max_x // 3
            if col0_width < 20:
                col0_width = 20
            other_col_count = ncols - 1
            if other_col_count < 1:
                other_col_count = 1
            remaining_width = max_x - col0_width
            each_col_width = remaining_width // other_col_count
            if each_col_width < 12:
                each_col_width = 12

            def get_col_width(ix):
                return col0_width if ix == 0 else each_col_width

            try:

                # Instructions lines
                stdscr.addstr(0, 0, "=== nifti-convert: Convert NIfTI to BIDS for QSMxT ==="[:max_x])
                stdscr.addstr(1, 0, "INSTRUCTIONS: Fill out the data table below to complete the conversion to BIDS."[:max_x])

                stdscr.addstr(3, 0, "CONTROLS: ARROWS=move; ESC=done; TAB=toggle full paths; DEL=clear / copy above cell"[:max_x])
                stdscr.addstr(4, 0, "REGULAR EXPRESSIONS: You can use the regex row to autopopulate columns based on filenames."[:max_x])

                # Status message
                stdscr.addstr(6, 0, f"STATUS: {status[1]}"[:max_x], status[0])

                if row_idx == -1:
                    help_text = "HELP <REGEX row>: Edit a pattern for each column. Press ENTER to apply."[:max_x]
                else:
                    current_col = columns[col_idx]
                    help_text = f"HELP <{current_col}> {column_help.get(current_col, '')}"[:max_x]
                stdscr.addstr(7, 0, help_text[:max_x])

                # Blank line 4
                regex_header_y = 9
                # Regex header
                regex_header_str = ""
                for ci, cname in enumerate(columns):
                    w = get_col_width(ci)
                    regex_header_str += cname[:w].ljust(w) + " "
                stdscr.addstr(regex_header_y, 0, regex_header_str[:max_x])

                # Regex row
                regex_row_y = regex_header_y + 1
                row_output = ""
                for ci, cname in enumerate(columns):
                    w = get_col_width(ci)
                    if ci == 0:
                        cell_txt = "[NoRegex]"
                    else:
                        cell_txt = regex_map[cname][:w]

                    if row_idx == -1 and col_idx == ci:
                        stdscr.attron(curses.A_REVERSE)
                        stdscr.addstr(regex_row_y, len(row_output), cell_txt.ljust(w))
                        stdscr.attroff(curses.A_REVERSE)
                    else:
                        stdscr.addstr(regex_row_y, len(row_output), cell_txt.ljust(w))
                    row_output += cell_txt.ljust(w) + " "

                # Blank line 7
                table_header_y = regex_row_y + 2
                # Table header
                table_header_str = ""
                for ci, cname in enumerate(columns):
                    w = get_col_width(ci)
                    table_header_str += cname[:w].ljust(w) + " "
                stdscr.addstr(table_header_y, 0, table_header_str[:max_x])

                data_start_y = table_header_y + 1

                # Data rows
                for r_i in range(nrows):
                    yy = data_start_y + r_i
                    if yy >= max_y:
                        break

                    col_x = 0
                    for ci, cname in enumerate(columns):
                        w = get_col_width(ci)
                        val = ""
                        if ci == 0:
                            # NIfTI_Path
                            if show_full_path:
                                val = str(session.loc[r_i, "NIfTI_Path_Full"])
                            else:
                                val = os.path.basename(str(session.loc[r_i, "NIfTI_Path_Full"]))
                        else:
                            raw_val = session.loc[r_i, cname]
                            val = "" if pd.isnull(raw_val) else str(raw_val)

                        val = val[:w]

                        # highlight if this is the current cell
                        if (r_i == row_idx) and (ci == col_idx) and (row_idx >= 0):
                            stdscr.attron(curses.A_REVERSE)
                            stdscr.addstr(yy, col_x, val.ljust(w))
                            stdscr.attroff(curses.A_REVERSE)
                        else:
                            stdscr.addstr(yy, col_x, val.ljust(w))
                        col_x += w + 1
            except curses.error:
                stdscr.clear()
                stdscr.addstr(0, 0, "Window too small!", error)
                stdscr.addstr(1, 0, "Please resize the window to be larger.", error)
                stdscr.refresh()
                stdscr.getch()
                continue

            stdscr.refresh()
            key = stdscr.getch()

            if key == 27:  # ESC
                valid, errors = validate_data(session)
                if not valid:
                    row_idx, col_idx, msg = errors[0]
                    status = (error, msg)
                else:
                    return

            elif key in (curses.KEY_ENTER, 10, 13):
                # ENTER
                if row_idx == -1:
                    # apply regex for col_idx if not 0
                    if col_idx == 0:
                        status = (warning, "Cannot apply regex to NIfTI_Path.")
                    else:
                        c_name = columns[col_idx]
                        valid, msg = apply_regex_to_filenames(c_name)
                        if valid:
                            status = (success, msg)
                        else:
                            status = (error, msg)
                else:
                    row_idx = min(nrows - 1, row_idx + 1)

            elif key == curses.KEY_UP:
                if row_idx == -1:
                    # can't go above regex row
                    pass
                elif row_idx == 0:
                    # moving from first data row to regex row
                    row_idx = -1
                else:
                    row_idx -= 1

            elif key == curses.KEY_DOWN:
                if row_idx == -1:
                    # move from regex row to first data row
                    row_idx = 0
                else:
                    row_idx = min(nrows - 1, row_idx + 1)

            elif key == curses.KEY_LEFT:
                col_idx = max(0, col_idx - 1)

            elif key == curses.KEY_RIGHT:
                col_idx = min(ncols - 1, col_idx + 1)

            elif key == 9:  # TAB => toggle path mode
                show_full_path = not show_full_path

            elif 32 <= key <= 126:
                # typed a normal ASCII char
                if row_idx == -1:
                    # editing regex
                    if col_idx != 0:
                        c_name = columns[col_idx]
                        regex_map[c_name] += chr(key)
                else:
                    # editing table
                    if col_idx != 0:  # NIfTI_Path is read-only
                        c_name = columns[col_idx]
                        curr_val = session.loc[row_idx, c_name]
                        if pd.isnull(curr_val):
                            curr_val = ""
                        session.loc[row_idx, c_name] = curr_val + chr(key)
                        changes = True

            elif key in (curses.KEY_BACKSPACE, 127):
                # BACKSPACE => remove last char if not empty, else copy from above
                if row_idx == -1 and col_idx != 0:
                    # editing regex pattern
                    c_name = columns[col_idx]
                    curr = regex_map[c_name]
                    if curr:
                        regex_map[c_name] = curr[:-1]
                    # if it's empty already, do nothing for regex row
                elif row_idx >= 0 and col_idx != 0:
                    # editing table
                    c_name = columns[col_idx]
                    curr_val = session.loc[row_idx, c_name]
                    if pd.isnull(curr_val) or str(curr_val).strip() == "":
                        curr_val = None
                    elif curr_val == None:
                        # copy from above if possible
                        copy_from_above_if_exists(row_idx, c_name)
                    else:
                        # remove last character
                        session.loc[row_idx, c_name] = curr_val[:-1]
                        if str(session.loc[row_idx, c_name]).strip() == "":
                            session.loc[row_idx, c_name] = None
                    changes = True

            elif key == curses.KEY_DC:
                # DELETE => clear if not empty, else copy from above
                if row_idx == -1 and col_idx != 0:
                    # editing regex
                    c_name = columns[col_idx]
                    curr = regex_map[c_name]
                    if curr == None:
                        # do nothing for regex row if it's already empty
                        pass
                    else:
                        # clear it
                        regex_map[c_name] = None
                elif row_idx >= 0 and col_idx != 0:
                    c_name = columns[col_idx]
                    curr_val = session.loc[row_idx, c_name]
                    if pd.isnull(curr_val):
                        curr_val = None
                    if curr_val == None:
                        # already empty => copy from above
                        copy_from_above_if_exists(row_idx, c_name)
                    else:
                        # clear
                        session.loc[row_idx, c_name] = None
                changes = True

            if changes:
                valid, errors = validate_data(session)
                if not valid:
                    _, _, msg = errors[0]
                    status = (error, msg)
                else:
                    status = (success, "Valid")
            # else ignore other keys

    curses.wrapper(curses_ui)
    return session

def nifti_convert(nifti_dir, output_dir):
    logger = make_logger()

    session = load_nifti_session(nifti_dir)
    session.reset_index(drop=True, inplace=True)
    # Sort by some BIDS fields if they exist, to get a consistent order
    sort_cols = [
        "sub", "ses", "acq", "run", "echo", "part",
        "suffix", "MagneticFieldStrength", "EchoTime", "NIfTI_Path"
    ]
    # Only use columns that exist
    use_sort_cols = [c for c in sort_cols if c in session.columns]
    session.sort_values(by=use_sort_cols, inplace=True)
    logger.log(LogLevel.INFO.value, f"Found {len(session)} NIfTI files.")

    session = interactive_table(session)

    # for each row in the dataframe, move the file to the new location
    for i in range(len(session)):
        row = session.loc[i]
        new_name = f"sub-{row['sub']}"
        if pd.notnull(row["ses"]):
            new_name += f"_ses-{row['ses']}"
        if pd.notnull(row["acq"]):
            new_name += f"_acq-{row['acq']}"
        if pd.notnull(row["run"]):
            new_name += f"_run-{row['run']}"
        if pd.notnull(row["echo"]):
            new_name += f"_echo-{row['echo']}"
        if pd.notnull(row["part"]):
            new_name += f"_part-{row['part']}"
        new_name += f"_{row['suffix']}.nii"

        anat_dir = os.path.join(output_dir, f"sub-{row['sub']}")
        if pd.notnull(row["ses"]):
            anat_dir = os.path.join(anat_dir, f"ses-{row['ses']}")
        anat_dir = os.path.join(anat_dir, "anat")

        os.makedirs(anat_dir, exist_ok=True)

        logger.log(LogLevel.INFO.value, f"Copying {row['NIfTI_Path_Full']} to {os.path.join(anat_dir, new_name)}")
        shutil.copy(row["NIfTI_Path_Full"], os.path.join(anat_dir, new_name))
        json_path = row['JSON_Path']
        shutil.copy(json_path, os.path.join(anat_dir, new_name.replace(".nii.gz", ".nii").replace(".nii", ".json")))

def main():
    parser = argparse.ArgumentParser(
        description="QSMxT niftiConvert with extended curses UI, multiline regex row, toggling path view, and field copying.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        'input_dir',
        help='Input NIfTI directory.'
    )

    parser.add_argument(
        'output_dir',
        help='Output BIDS directory.'
    )

    args = parser.parse_args()

    args.input_dir = os.path.abspath(args.input_dir)
    args.output_dir = os.path.abspath(args.output_dir)
    os.makedirs(args.output_dir, exist_ok=True)

    logger = make_logger(
        logpath=os.path.join(args.output_dir, f"log_{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}.txt"),
        printlevel=LogLevel.INFO,
        writelevel=LogLevel.INFO,
        warnlevel=LogLevel.WARNING,
        errorlevel=LogLevel.ERROR
    )

    logger.log(LogLevel.INFO.value, f"Running QSMxT {get_qsmxt_version()}")
    logger.log(LogLevel.INFO.value, f"Command: {str.join(' ', sys.argv)}")
    logger.log(LogLevel.INFO.value, f"Python interpreter: {sys.executable}")

    diff = get_diff()
    if diff:
        logger.log(LogLevel.WARNING.value, f"Working directory not clean! Writing diff to {os.path.join(args.output_dir, 'diff.txt')}...")
        with open(os.path.join(args.output_dir, "diff.txt"), "w") as diff_file:
            diff_file.write(diff)

    with open(os.path.join(args.output_dir, "references.txt"), 'w', encoding='utf-8') as f:
        f.write(f"QSMxT: {get_qsmxt_version()}")
        f.write(f"\nRun command: {str.join(' ', sys.argv)}")
        f.write(f"\nPython interpreter: {sys.executable}")
        f.write("\n\n == References ==\n")
        f.write("\n - Stewart AW, Bollmann S, et al. QSMxT/QSMxT. GitHub; 2022. https://github.com/QSMxT/QSMxT")
        f.write("\n - Gorgolewski KJ, Auer T, Calhoun VD, et al. The brain imaging data structure, a format for organizing outputs of neuroimaging experiments. Sci Data. 2016;3(1):160044.\n")

    nifti_convert(args.input_dir, args.output_dir)
    script_exit()

if __name__ == "__main__":
    main()



import numpy as np

from nipype.pipeline.engine import Workflow
from nipype.interfaces.utility import IdentityInterface, Function

from qsmxt.interfaces import nipype_interface_masking as masking
from qsmxt.interfaces import nipype_interface_erode as erode
from qsmxt.interfaces import nipype_interface_bet2 as bet2
from qsmxt.interfaces import nipype_interface_hdbet as hdbet
from qsmxt.interfaces import nipype_interface_phaseweights as phaseweights
from qsmxt.interfaces import nipype_interface_twopass as twopass
from qsmxt.interfaces import nipype_interface_makehomogeneous as makehomogeneous
from qsmxt.interfaces import nipype_interface_combinemagnitude as combinemagnitude

from qsmxt.scripts.qsmxt_functions import gen_plugin_args, create_node

def masking_workflow(run_args, mask_available, magnitude_available, qualitymap_available, fill_masks, add_bet, use_maps, name, dimensions_phase, bytepix_phase, num_echoes, index):

    wf = Workflow(name=f"{name}_workflow")

    slurm_account = run_args.slurm[0] if run_args.slurm and len(run_args.slurm) else None
    slurm_partition = run_args.slurm[1] if run_args.slurm and len(run_args.slurm) > 1 else None

    n_inputs = create_node(
        interface=IdentityInterface(
            fields=['phase', 'quality_map', 'magnitude', 'mask', 'TE']
        ),
        name='masking_inputs'
    )

    n_outputs = create_node(
        interface=IdentityInterface(
            fields=['mask', 'threshold', 'quality_map']
        ),
        name='masking_outputs'
    )

    if not mask_available:

        # determine whether bet will be used this run
        bet_used = magnitude_available and (
             run_args.masking_algorithm == 'bet'
             or run_args.add_bet
             or run_args.filling_algorithm == 'bet')
        bet_this_run = bet_used and (fill_masks or (run_args.mask_erosions and run_args.masking_algorithm == 'threshold' and not fill_masks))

        # combine magnitude if necessary
        if magnitude_available and run_args.combine_phase:
            n_combine_magnitude_mem = (np.prod(dimensions_phase) * 8) / (1024 ** 3) * num_echoes * 4
            n_combine_magnitude = create_node(
                interface=combinemagnitude.CombineMagnitudeInterface(),
                name='nibabel-numpy_combine-magnitude',
                mem_gb=n_combine_magnitude_mem
            )
            n_combine_magnitude.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=slurm_account,
                pbs_account=run_args.pbs,
                slurm_partition=slurm_partition,
                name="comb-mag",
                time="01:00:00",
                mem_gb=n_combine_magnitude_mem
            )
            wf.connect([
                (n_inputs, n_combine_magnitude, [('magnitude', 'magnitude')])
            ])

        # get first phase image if necessary
        if run_args.combine_phase and run_args.masking_input == 'phase':
            n_getfirst_phase = create_node(
                interface=Function(
                    input_names=['phase', 'TE', 'is_list'],
                    output_names=['phase', 'TE'],
                    function=lambda phase, TE, is_list: [phase[0], TE[0]] if is_list else [phase, TE]
                ),
                name='func_get-first-phase'
            )
            n_getfirst_phase.inputs.is_list = True
            wf.connect([
                (n_inputs, n_getfirst_phase, [('phase', 'phase')]),
                (n_inputs, n_getfirst_phase, [('TE', 'TE')])
            ])

        # correct magnitude if necessary
        if run_args.inhomogeneity_correction and (bet_this_run or run_args.masking_input == 'magnitude'):
            mn_inhomogeneity_correction = create_node(
                interface=makehomogeneous.MakeHomogeneousInterface(),
                iterfield=['magnitude'],
                name='mrt_correct-inhomogeneity',
                is_map=use_maps
            )

            if run_args.combine_phase:
                wf.connect([
                    (n_combine_magnitude, mn_inhomogeneity_correction, [('magnitude_combined', 'magnitude')]),
                ])
            else:
                wf.connect([
                    (n_inputs, mn_inhomogeneity_correction, [('magnitude', 'magnitude')])
                ])

        # do phase weights if necessary
        if run_args.masking_algorithm == 'threshold' and run_args.masking_input == 'phase':
            mn_phaseweights_mem = 2.80122 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 0.95 # DONE
            mn_phaseweights_threads = 1
            if qualitymap_available:
                mn_phaseweights = create_node(
                    interface=IdentityInterface(['quality_map']),
                    name='romeo-voxelquality',
                    n_procs=mn_phaseweights_threads,
                    mem_gb=mn_phaseweights_mem,
                )
                wf.connect([(n_inputs, mn_phaseweights, [('quality_map', 'quality_map')])])
            else:
                mn_phaseweights = create_node(
                    interface=phaseweights.RomeoMaskingInterface(),
                    iterfield=['phase', 'magnitude'] if magnitude_available else ['phase'],
                    name='romeo-voxelquality',
                    mem_gb=mn_phaseweights_mem,
                    n_procs=mn_phaseweights_threads,
                    is_map=use_maps
                )
                mn_phaseweights.inputs.weight_type = "grad+second"
                wf.connect([
                    (n_getfirst_phase if (run_args.combine_phase and run_args.masking_input == 'phase') else n_inputs, mn_phaseweights, [('phase', 'phase')]),
                    (n_getfirst_phase if (run_args.combine_phase and run_args.masking_input == 'phase') else n_inputs, mn_phaseweights, [('TE', 'TEs' if use_maps else 'TE')]),
                    (mn_phaseweights, n_outputs, [('quality_map', 'quality_map')])
                ])
                if magnitude_available:
                    mn_phaseweights.inputs.weight_type = "grad+second+mag_weight+mag_coherence"
                    if run_args.combine_phase:
                        wf.connect([
                            (n_combine_magnitude, mn_phaseweights, [('magnitude_combined', 'magnitude')])
                        ])
                    else:
                        wf.connect([
                            (n_inputs, mn_phaseweights, [('magnitude', 'magnitude')])
                        ])
            
            mn_phaseweights.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=slurm_account,
                pbs_account=run_args.pbs,
                slurm_partition=slurm_partition,
                name="voxelquality",
                time="01:00:00",
                mem_gb=mn_phaseweights_mem,
                num_cpus=mn_phaseweights_threads
            )

        # do bet mask if necessary
        if bet_this_run:
            bet_threads = min(8, run_args.n_procs) if run_args.multiproc else 8
            bet_mem = (np.prod(dimensions_phase) * bytepix_phase) / (1024 ** 3) * 10
            mn_bet = create_node(
                interface=bet2.Bet2Interface(fractional_intensity=run_args.bet_fractional_intensity),
                iterfield=['in_file'],
                name='fsl-bet',
                mem_gb=bet_mem,
                n_procs=bet_threads,
                is_map=use_maps
            )
            mn_bet.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=slurm_account,
                pbs_account=run_args.pbs,
                slurm_partition=slurm_partition,
                name="bet",
                time="01:00:00",
                mem_gb=bet_mem,
                num_cpus=bet_threads
            )
            if run_args.inhomogeneity_correction:
                wf.connect([
                    (mn_inhomogeneity_correction, mn_bet, [('magnitude_corrected', 'in_file')])
                ])
            elif run_args.combine_phase:
                wf.connect([
                    (n_combine_magnitude, mn_bet, [('magnitude_combined', 'in_file')])
                ])
            else:
                wf.connect([
                    (n_inputs, mn_bet, [('magnitude', 'in_file')])
                ])

            # erode bet mask
            mn_bet_erode = create_node(
                interface=erode.ErosionInterface(
                    num_erosions=run_args.mask_erosions[index % len(run_args.mask_erosions)] if run_args.mask_erosions else 0
                ),
                iterfield=['in_file'],
                name='scipy_numpy_nibabel_bet_erode',
                is_map=use_maps
            )
            wf.connect([
                (mn_bet, mn_bet_erode, [('mask', 'in_file')])
            ])

            # output eroded bet mask if necessary
            if run_args.masking_algorithm == 'bet' or (fill_masks and run_args.filling_algorithm == 'bet'):
                wf.connect([
                    (mn_bet_erode, n_outputs, [('out_file', 'mask')])
                ])

        # do threshold masking if necessary
        if run_args.masking_algorithm == 'threshold' and not (fill_masks and run_args.filling_algorithm == 'bet'):
            n_threshold_masking = create_node(
                interface=masking.MaskingInterface(
                    threshold_algorithm='otsu' or run_args.threshold_algorithm,
                    threshold_algorithm_factor=run_args.threshold_algorithm_factor[index % len(run_args.threshold_algorithm_factor)],
                    fill_masks=fill_masks,
                    mask_suffix=name,
                    num_erosions=run_args.mask_erosions[index % len(run_args.mask_erosions)] if run_args.mask_erosions else 0,
                    filling_algorithm=run_args.filling_algorithm
                ),
                name='scipy_numpy_nibabel_threshold-masking'
                # inputs : ['in_files']
            )
            if run_args.threshold_value:
                n_threshold_masking.inputs.threshold = run_args.threshold_value[index % len(run_args.threshold_value)]
            if bet_this_run:
                wf.connect([
                    (mn_bet_erode, n_threshold_masking, [('out_file', 'bet_masks')])
                ])
            if run_args.masking_input == 'phase':
                wf.connect([
                    (mn_phaseweights, n_threshold_masking, [('quality_map', 'in_files')])
                ])
            elif run_args.masking_input == 'magnitude':
                if run_args.inhomogeneity_correction:
                    wf.connect([
                        (mn_inhomogeneity_correction, n_threshold_masking, [('magnitude_corrected', 'in_files')])
                    ])
                elif run_args.combine_phase:
                    wf.connect([
                        (n_combine_magnitude, n_threshold_masking, [('magnitude_combined', 'in_files')])
                    ])
                else:
                    wf.connect([
                        (n_inputs, n_threshold_masking, [('magnitude', 'in_files')])
                    ])
            if not add_bet:
                wf.connect([
                    (n_threshold_masking, n_outputs, [('mask', 'mask')])
                ])
            else:
                mn_mask_plus_bet = create_node(
                    interface=twopass.TwopassNiftiInterface(),
                    name='numpy_nibabel_mask-plus-bet',
                    iterfield=['in_file', 'in_filled'],
                    is_map=use_maps
                )
                wf.connect([
                    (n_threshold_masking, mn_mask_plus_bet, [('mask', 'in_file')]),
                    (mn_bet_erode, mn_mask_plus_bet, [('out_file', 'in_filled')]),
                    (mn_mask_plus_bet, n_outputs, [('out_file', 'mask')])
                ])

    # outputs
    if mask_available:
        wf.connect([
            (n_inputs, n_outputs, [('mask', 'mask')]),
        ])
    else:
        if run_args.masking_algorithm == 'threshold' and not (fill_masks and run_args.filling_algorithm == 'bet'):
            wf.connect([
                (n_threshold_masking, n_outputs, [('threshold', 'threshold')])
            ])

    return wf






def get_string(prompt, default=None):
    user_in = input(prompt)
    if not user_in: return default
    return user_in

def get_option(prompt, options, default=None):
    while True:
        user_in = input(prompt)
        if not user_in:
            return default
        if user_in in options:
            return user_in

def get_num(prompt, default=None, min_val=None, max_val=None, dtype=float):
    while True:
        user_in = input(prompt)
        if not user_in: return default
        try:
            user_in = float(user_in)
        except ValueError:
            continue
        try:
            if dtype(user_in) == user_in:
                user_in = dtype(user_in)
        except ValueError:
            continue
        if min_val and user_in < min_val: continue
        if max_val and user_in > max_val: continue
        return user_in

def get_nums(prompt, default=None, min_val=None, max_val=None, min_n=None, max_n=None, dtype=float):
    while True:
        user_in = input(prompt)
        if not user_in: return default

        # split input
        user_in = user_in.replace('[', '').replace(']', '').replace('(', '').replace(')', '').replace(',', ' ').split()

        # check for n
        if min_n and len(user_in) < min_n: continue
        if max_n and len(user_in) > max_n: continue

        # convert to numeric and check range
        success = True
        for i in range(len(user_in)):
            try:
                user_in[i] = float(user_in[i])
            except ValueError:
                success = False
                break
            if min_val and user_in[i] < min_val: continue
            if max_val and user_in[i] > max_val: continue

        # convert to desired type (e.g. int)
        if not success: continue
        if dtype != float:
            for i in range(len(user_in)):
                try:
                    if dtype(user_in[i]) == user_in[i]:
                        user_in[i] = dtype(user_in[i])
                except ValueError:
                    success = False
                    break
            if not success: continue

        return user_in
        





import os
import shutil
import json
import site
import pkg_resources
import subprocess
import psutil
import math
from qsmxt.scripts.logger import LogLevel, make_logger
from qsmxt.scripts.sys_cmd import sys_cmd
from nipype.pipeline.engine import Node, MapNode

def create_node(interface, name, n_procs=1, mem_gb=2, iterfield=None, is_map=False, **kwargs):
    logger = make_logger('main')
    mem_gb = round(mem_gb, 3)
    mem_avail = round(psutil.virtual_memory().available / (1024 ** 3) * 0.90, 3)
    logger.log(LogLevel.DEBUG.value, f"Node {name} has requested {mem_gb} GB.")
    if mem_gb < 2:
        mem_gb = 2
    if mem_gb > mem_avail:
        logger.log(LogLevel.WARNING.value, f"Node {name} has requested {mem_gb} GB of memory, which is greater than the available memory {mem_avail} GB! Segmentation faults may occur.")
        mem_gb = mem_avail
    if is_map:
        return MapNode(interface=interface, name=name, iterfield=iterfield, n_procs=n_procs, mem_gb=mem_gb, **kwargs)
    else:
        return Node(interface=interface, name=name, n_procs=n_procs, mem_gb=mem_gb, **kwargs)

def get_qsmxt_dir():
    path = os.path.abspath(__file__)

    # Check if the path contains 'site-packages'; if so, it's installed as a package
    if 'site-packages' in path:
        return os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(path))), 'qsmxt')
    else:
        # Return the directory up one level from the current file if running from source
        return os.path.dirname(os.path.dirname(path))

def is_editable_package(package_name):
    """
    Determine if a package was installed in "editable" mode.
    
    :param package_name: The name of the package.
    :return: True if the package was installed in editable mode, False otherwise.
    """
    
    # Get the site-packages directory
    site_packages = site.getsitepackages()[0]
    
    # Look for the package's metadata directory
    for item in os.listdir(site_packages):
        if item.startswith(package_name) and item.endswith(".egg-link"):
            return True
        if item.startswith(package_name) and item.endswith(".dist-info"):
            dist_info_dir = os.path.join(site_packages, item)
            direct_url_path = os.path.join(dist_info_dir, "direct_url.json")
            
            # If direct_url.json exists, parse it and check for "editable"
            if os.path.exists(direct_url_path):
                with open(direct_url_path, 'r') as f:
                    data = json.load(f)
                    return data.get("editable", False)
    
    return False

def get_qsmxt_version():
    return f"{pkg_resources.get_distribution('qsmxt').version}" + (" (linked installation)" if is_editable_package('qsmxt') else "")

def get_qsm_premades(pipeline_file=None):
    with open(f"{os.path.join(get_qsmxt_dir(), 'qsm_pipelines.json')}", "r") as fh:
        premades = json.load(fh)

    if pipeline_file:
        with open(pipeline_file, "r") as fh:
            user_premades = json.load(fh)
        premades.update(user_premades)

    return premades

def extend_fname(original_path, append_text, ext=None, out_dir=None):
    out_dir = out_dir or os.path.split(original_path)[0]
    original_fname = os.path.split(original_path)[1].split('.')[0]
    original_ext = ".".join(os.path.split(original_path)[1].split('.')[1:])
    return os.path.join(out_dir, f"{original_fname}{append_text}.{ext or original_ext}")

def get_fname(path, include_path=True):
    path_noext = ".".join(path.split('.')[:-1])
    if include_path: return path_noext
    return os.path.split(path_noext)[-1]

def print_qsm_premades(pipeline_file):
    premades = get_qsm_premades(pipeline_file)
    print("=== Premade pipelines ===")
    for key, value in premades.items():
        print(f"{key}", end="")
        if "description" in value:
            print(f": {value['description']}")
        else:
            print()

def get_container_version(check_path=True):
    if os.environ.get('APPTAINER_NAME') and 'qsmxt' in os.environ.get('APPTAINER_NAME'):
        return f"{os.environ.get('APPTAINER_NAME')} (apptainer)"
    if os.path.exists("/README.md"):
        with open("/README.md", 'r') as readme_handle:
            lines = readme_handle.readlines()
        for line in lines:
            if '## qsmxt/' in line:
                version = line.split("/")[1].split(" ")[0].strip()
                return f"{version}"
    if check_path and shutil.which("qsmxt_version.py"):
        return sys_cmd("qsmxt_version.py --container_only", print_output=False, print_command=False)
    return "unknown"

def is_git_repo(directory):
    cmd = f"git --git-dir {os.path.join(directory, '.git')} rev-parse"
    try:
        sys_cmd(cmd, print_output=False, print_command=False, raise_exception=True)
        return True
    except subprocess.CalledProcessError:
        return False

def get_diff():
    qsmxt_dir = get_qsmxt_dir()
    if is_git_repo(qsmxt_dir):
        diff = sys_cmd(f"git --git-dir {os.path.join(qsmxt_dir, '.git')} --work-tree {qsmxt_dir} diff", print_command=False, print_output=False)
        return f"{diff}\n" if diff else ""
    else:
        return ""

def gen_plugin_args(pbs_account="", slurm_account="", plugin_args={}, time="00:30:00", num_cpus=1, mem_gb=5, name="QSMxT", slurm_partition=None):
    mem_gb = math.ceil(mem_gb)
    plugin_args['sbatch_args'] = f"--account={slurm_account} {f'--partition {slurm_partition}' if slurm_partition else ''} --job-name={name} --time={time} --ntasks=1 --cpus-per-task={num_cpus} --mem={mem_gb}gb"
    plugin_args['qsub_args'] = f'-A {pbs_account} -N {name} -l walltime={time} -l select=1:ncpus={num_cpus}:mem={mem_gb}gb'
    return plugin_args
    



#!/usr/bin/env python

import nibabel as nib
import numpy as np
import argparse
import os
from qsmxt.scripts.qsmxt_functions import extend_fname
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, traits

def reference_susceptibility(in_qsm, in_seg=None, in_seg_values=None, out_qsm=None):
    qsm_nii = nib.load(in_qsm)
    qsm = np.array(qsm_nii.get_fdata(), dtype=np.float64)

    full_mask = abs(qsm) >= 5e-5
    if in_seg and in_seg_values:
        seg = np.array(nib.load(in_seg).get_fdata(), dtype=np.int64)
        ref_mask = np.zeros_like(seg, dtype=np.int64)
        for value in in_seg_values:
            ref_mask = np.logical_or(ref_mask, seg.astype(np.int64) == np.int64(value))
    else:
        ref_mask = full_mask
    
    nib.save(nib.Nifti1Image(ref_mask, qsm_nii.affine, qsm_nii.header), extend_fname(in_qsm, "_ref_mask", out_dir=os.getcwd()))

    qsm[full_mask] = qsm[full_mask] - np.mean(qsm[ref_mask == 1])

    out_qsm = out_qsm or extend_fname(in_qsm, "_ref", out_dir=os.getcwd())
    nib.save(nib.Nifti1Image(qsm, qsm_nii.affine, qsm_nii.header), out_qsm)
    return out_qsm


class ReferenceQSMInputSpec(BaseInterfaceInputSpec):
    in_qsm = File(mandatory=True, exists=True)
    in_seg = File(mandatory=False, exists=True)
    in_seg_values = traits.Either(traits.ListInt, None, mandatory=False, default=None)


class ReferenceQSMOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class ReferenceQSMInterface(SimpleInterface):
    input_spec = ReferenceQSMInputSpec
    output_spec = ReferenceQSMOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = reference_susceptibility(
            in_qsm=self.inputs.in_qsm,
            in_seg=self.inputs.in_seg,
            in_seg_values=self.inputs.in_seg_values
        )
        return runtime


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Reference a susceptibility map to the mean of the values within a segmented region.")
    parser.add_argument("in_qsm", help="Path to the input susceptibility map.")
    parser.add_argument("--in_seg", help="Optional path to the input segmentation. By default, any non-zero value will be considered a part of the mask.")
    parser.add_argument("--in_seg_values", default=[1], nargs='+', help="Optional values to use for the segmentation mask.")
    parser.add_argument("out_qsm", help="Path to save the output referenced susceptibility map.")
    args = parser.parse_args()

    reference_susceptibility(args.in_qsm, args.in_seg, args.in_seg_values, args.out_qsm)




#!/usr/bin/env python3

import argparse
import os
import sys
import subprocess
import json
import datetime
import re
import curses
import shutil
import time
import pandas as pd
import fnmatch

from tabulate import tabulate

from dicompare import load_dicom_session, load_nifti_session

from qsmxt.scripts.qsmxt_functions import get_qsmxt_version, get_diff
from qsmxt.scripts.logger import LogLevel, make_logger, show_warning_summary 
from qsmxt.scripts.nii_fix_ge import fix_ge_polar, fix_ge_complex

from collections import Counter

def sys_cmd(cmd):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"Running command: '{cmd}'")
        
    process = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout_byte = process.stdout
    stderr_byte = process.stderr

    stdout_str = stdout_byte.decode('UTF-8')
    stderr_str = str(stderr_byte.decode('UTF-8'))
    return_code = process.returncode
    
    if stdout_str:
        logger.log(LogLevel.DEBUG.value, f"Command output: '{stdout_str}'", end="")

    if return_code:
        logger.log(LogLevel.WARNING.value, f"Command '{cmd}' returned error {return_code}: '{stderr_str}'")
    
    return return_code

def load_json(path):
    with open(path, encoding='utf-8') as f:
        j = json.load(f)
    return j

def rename(old, new, always_show=False):
    if always_show or not sys.__stdin__.isatty():
        logger = make_logger()
        logger.log(LogLevel.INFO.value, f'Renaming {old} -> {new}')
    os.makedirs(os.path.split(new)[0], exist_ok=True)
    os.rename(old, new)

def clean(data): 
    cleaned = re.sub(r'[^a-zA-Z0-9]', '', data).lower()
    if data.startswith('sub-'):
        return f'sub-{cleaned[3:]}'
    elif data.startswith('ses-'):
        return f'ses-{cleaned[3:]}'
    return cleaned

def get_folders_in(folder, full_path=False):
    folders = list(filter(os.path.isdir, [os.path.join(folder, d) for d in os.listdir(folder)]))
    if full_path:
        return folders
    folders = [os.path.split(x)[1] for x in folders]
    return folders

def find_and_autoassign_qsm_pairs(row_data):
    """
    Automatically identifies exactly one pair for QSM per (Count, NumEchoes),
    picking the pair (Mag+Phase or Real+Imag) with the smallest difference
    in SeriesNumber. Priority:
       1) Mag + Phase
       2) Real + Imag

    Returns None if at least one valid pair is found, or a message if none is found.
    """

    from collections import defaultdict, Counter

    # Group rows by (Count, NumEchoes). If you only want to group by Count, remove NumEchoes
    grouped = defaultdict(list)
    for r in row_data:
        cval = (r["Count"], r["NumEchoes"])
        grouped[cval].append(r)

    any_pair_found = False

    for (count_val, echo_val), group_rows in grouped.items():
        # Mark everything Skip first
        for row in group_rows:
            row["Type"] = "Skip"

        # Gather potential M/P or R/I based on ImageType
        mag_list   = []
        phase_list = []
        real_list  = []
        imag_list  = []

        def classify(row):
            """Return 'mag', 'phase', 'real', or 'imag' if the row qualifies, else None."""
            # Consider row['ImageType'] (tuple?)
            # We'll unify to uppercase strings:
            image_type = row.get("ImageType", ())
            if isinstance(image_type, (tuple, list)):
                image_type_up = [x.upper() for x in image_type]
            else:
                image_type_up = [str(image_type).upper()]

            # Some helper for checking the row's ImageType
            def has_any(substrings, container):
                return any(s in container for s in substrings)

            # Priority #1: mag
            if has_any(["MAG", "M"], image_type_up):
                return "mag"
            # Priority #2: phase
            if has_any(["PHASE", "P"], image_type_up):
                return "phase"
            # Priority #3: real
            if has_any(["REAL"], image_type_up):
                return "real"
            # Priority #4: imag (imaginary)
            if has_any(["IMAG", "IMAGINARY"], image_type_up):
                return "imag"
            return None

        # Classify each row
        for row in group_rows:
            ctype = classify(row)
            if ctype == "mag":
                mag_list.append(row)
            elif ctype == "phase":
                phase_list.append(row)
            elif ctype == "real":
                real_list.append(row)
            elif ctype == "imag":
                imag_list.append(row)
            # else remains "Skip"

        # Next, we pick exactly 1 mag+phase pair if possible
        # Among all possible pairs, choose the one with the SMALLEST difference in SeriesNumber
        def pick_closest_pair(listA, listB, labelA, labelB):
            """Return (rowA, rowB) that have the minimal |SeriesNumberA - SeriesNumberB|, or None."""
            if not listA or not listB:
                return None, None
            best_pair = None
            best_diff = float("inf")
            for a in listA:
                for b in listB:
                    # We'll assume row['SeriesNumber'] is an integer or numeric
                    # If it's missing, default to 999999 or 0, you can decide
                    #snA = a.get("SeriesNumber", 999999)
                    #snB = b.get("SeriesNumber", 999999)
                    # let's instead use the row idx
                    snA = group_rows.index(a)
                    snB = group_rows.index(b)
                    
                    diff = abs(snA - snB)
                    if diff < best_diff:
                        best_diff = diff
                        best_pair = (a, b)
            if best_pair:
                best_pair[0]["Type"] = labelA  # e.g. "Mag"
                best_pair[1]["Type"] = labelB  # e.g. "Phase"
                return best_pair
            return None, None

        # 1) Try picking best (mag, phase)
        mag_phase_pair = pick_closest_pair(mag_list, phase_list, "Mag", "Phase")
        if mag_phase_pair[0] and mag_phase_pair[1]:
            any_pair_found = True
        else:
            # 2) if no mag+phase, try real+imag
            real_imag_pair = pick_closest_pair(real_list, imag_list, "Real", "Imag")
            if real_imag_pair[0] and real_imag_pair[1]:
                any_pair_found = True
            else:
                # Nothing chosen => entire group remains "Skip"
                pass

    # If no pairs found in any group, do the repeated Count check
    if not any_pair_found:
        counts = Counter(r["Count"] for r in row_data)
        if not any(v >= 2 for v in counts.values()):
            return "Not suitable for QSM (no repeated Count)."
    return None

import re

def find_and_autoassign_t1w(row_data):
    """
    Automatically identifies exactly one series for T1w.

    Priority of matching (picks first match):
      1) If row_data has exactly one row, auto-assign T1w.
      2) "UNI?DEN" in ImageType or SeriesDescription (regex 'UNI.?DEN').
      3) "T1" in ImageType.
      4) "*t1w*" in SeriesDescription.
      5) "*t1*" in SeriesDescription.
      6) Otherwise, exclude any series with "T1 MAP" or "R1" or "R2" in
         ImageType or SeriesDescription, pick the first from the remainder if exactly one.
    
    If no candidate is found, all remain "Skip".
    """

    # Mark all as Skip
    for row in row_data:
        row["Type"] = "Skip"

    # If there's exactly one row, assign it T1w and return
    if len(row_data) == 1:
        row_data[0]["Type"] = "T1w"
        return

    # 1) Attempt "UNI?DEN" in ImageType or SeriesDescription
    # We'll define a quick function to check
    UNI_DEN_REGEX = re.compile(r'UNI.?DEN', re.IGNORECASE)

    for row in row_data:
        it_str = row.get("ImageType", [])
        sd_str = row.get("SeriesDescription", "")
        if any(UNI_DEN_REGEX.search(x) for x in it_str) or UNI_DEN_REGEX.search(sd_str):
            row["Type"] = "T1w"
            return  # done

    # 2) If no match yet, look for "T1" in ImageType
    for row in row_data:
        it_str = row.get("ImageType", [])
        if "T1" in it_str:
            row["Type"] = "T1w"
            return  # done

    # 3) If no match yet, SeriesDescription matching "*t1w*"
    for row in row_data:
        sd_str = row.get("SeriesDescription", "")
        if "T1W" in sd_str:
            row["Type"] = "T1w"
            return  # done

    # 4) If no match yet, SeriesDescription matching "*t1*"
    for row in row_data:
        sd_str = row.get("SeriesDescription", "")
        if "T1" in sd_str:
            row["Type"] = "T1w"
            return  # done

    def is_excluded_t1_map_r(row):
        it_str = row.get("ImageType", [])
        sd_str = row.get("SeriesDescription", "").upper()
        exclude_keywords = ["T1 MAP", "R1", "R2"]
        return any(k in it_str or k in sd_str for k in exclude_keywords)

    # build list of non-excluded
    candidates = [r for r in row_data if not is_excluded_t1_map_r(r)]
    
    # if candidates has one item exactly
    if len(candidates) == 1:
        candidates[0]["Type"] = "T1w"
        return

    # If we reached here, no T1w assigned (all remain Skip).
    return


def validate_qsm_rows(row_data):
    """
    Checks if at least one valid pair is assigned. 
    For each Count, allowed combinations:
      - 0 or 1 Mag, 0 or 1 Phase, or 
      - 0 or 1 Real, 0 or 1 Imag
    A pair is (Mag, Phase) or (Real, Imag) for that Count.
    If no pairs are found, or if there is any partial mismatch, returns an error message.
    """
    used_rows = [r for r in row_data if r['Type'] in ['Mag','Phase','Real','Imag']]
    if not used_rows:
        return (False, "No pairs have been selected.")
    
    # Group by Count
    grouped = {}
    for r in used_rows:
        cval = r['Count']
        if cval not in grouped:
            grouped[cval] = []
        grouped[cval].append(r)

    valid_pairs_found = False

    for cval, group in grouped.items():
        # Sort them by Type
        assigned_types = [g["Type"] for g in group]
        n_mag = assigned_types.count("Mag")
        n_pha = assigned_types.count("Phase")
        n_real = assigned_types.count("Real")
        n_imag = assigned_types.count("Imag")

        # If we have both Mag and Phase for this Count, that is one valid pair
        # If we also have Real and Imag, that's an additional valid pair
        # No mixing Real+Phase or Imag+Mag for the same row
        # More than one instance of the same label is not allowed here
        if n_mag > 1 or n_pha > 1 or n_real > 1 or n_imag > 1:
            return (False, f"More than one row labeled as Mag/Phase/Real/Imag for the same Count {cval}.")

        # If we have (Mag=1, Phase=1), that is a valid pair
        # If we have (Real=1, Imag=1), that is another valid pair
        # A partial mismatch is not allowed
        has_magpha = (n_mag == 1 and n_pha == 1)
        has_realimag = (n_real == 1 and n_imag == 1)
        partial_magpha = (n_mag == 1 and n_pha == 0) or (n_mag == 0 and n_pha == 1)
        partial_realimag = (n_real == 1 and n_imag == 0) or (n_real == 0 and n_imag == 1)

        if partial_magpha or partial_realimag:
            return (False, f"Partial assignment at Count {cval} is not allowed.")

        if has_magpha or has_realimag:
            valid_pairs_found = True
    
    if not valid_pairs_found:
        return (False, "No valid pairs were found in the assigned rows.")
    
    return (True, "")

def validate_t1w_rows(row_data):
    used = [r for r in row_data if r['Type'] == 'T1w']
    if len(used) < 1:
        return (False, "Must select at least one series for T1w.")
    return (True, "")

def detail_screen(stdscr, protocols, user_data):
    """
    Integrates the protocol loop that was previously in interactive_acquisition_selection.
    This function handles rendering and key interactions for each protocol in 'protocols'.
    """

    def do_validation(rows, acq_type):
        if acq_type == "QSM":
            ok, msg = validate_qsm_rows(rows)
            if not ok:
                # Place error message after the table
                error_line = TABLE_ROWS + len(rows) + 1
                stdscr.addstr(error_line, 0, f"ERROR: {msg}")
                return False
        elif acq_type == "T1w":
            ok, msg = validate_t1w_rows(rows)
            if not ok:
                error_line = TABLE_ROWS + len(rows) + 1
                stdscr.addstr(error_line, 0, f"ERROR: {msg}")
                return False
        return True

    protocol_idx = 0

    # Constants for screen lines
    INSTRUCTIONS_LINE_1 = 0
    INSTRUCTIONS_LINE_2 = 1
    PROTOCOL_NAME_LINE = 3
    ACQ_TYPE_LINE = 4
    TABLE_HEADER = 6
    TABLE_ROWS = 7

    detail_idx = PROTOCOL_NAME_LINE

    while True:
        stdscr.clear()

        # Two lines of instructions
        stdscr.addstr(INSTRUCTIONS_LINE_1, 0, "Use arrow keys: (↑/↓) move, (←/→) change, ENTER to confirm.")
        stdscr.addstr(INSTRUCTIONS_LINE_2, 0, "ESC exits or goes back to previous menu.")

        protocol = protocols[protocol_idx]
        rows = user_data[protocol]["Rows"]
        acq_type = user_data[protocol]["AcquisitionType"]

        # Protocol line
        proto_line = f"ProtocolName ({protocol_idx + 1} of {len(protocols)}): {protocol}"
        if detail_idx == PROTOCOL_NAME_LINE:
            stdscr.attron(curses.A_REVERSE)
            stdscr.addstr(PROTOCOL_NAME_LINE, 0, proto_line)
            stdscr.attroff(curses.A_REVERSE)
        else:
            stdscr.addstr(PROTOCOL_NAME_LINE, 0, proto_line)

        # Acquisition type line
        acq_line = f"Acquisition type: {acq_type}"
        if detail_idx == ACQ_TYPE_LINE:
            stdscr.attron(curses.A_REVERSE)
            stdscr.addstr(ACQ_TYPE_LINE, 0, acq_line)
            stdscr.attroff(curses.A_REVERSE)
        else:
            stdscr.addstr(ACQ_TYPE_LINE, 0, acq_line)

        # Table display
        headers = ["SeriesDescription", "ImageType", "Count", "NumEchoes", "Type"]
        table_data = [
            [r["SeriesDescription"], r["ImageType"], r["Count"], r["NumEchoes"], r["Type"]]
            for r in rows
        ]
            
        table_lines = tabulate(table_data, headers=headers, tablefmt="plain").split("\n")
        for i, line in enumerate(table_lines):
            row_y = TABLE_HEADER + i
            if i == 0:
                stdscr.addstr(row_y, 0, line)
            else:
                row_i = row_y
                if detail_idx == row_i:
                    stdscr.attron(curses.A_REVERSE)
                    stdscr.addstr(row_y, 0, line)
                    stdscr.attroff(curses.A_REVERSE)
                else:
                    stdscr.addstr(row_y, 0, line)

        # Run basic validation for display feedback
        valid = do_validation(rows, acq_type)

        stdscr.refresh()
        key = stdscr.getch()

        # If ENTER
        if key == 10:
            # Check all protocols for validity
            invalid_protocol = None
            for prot in protocols:
                test_acq_type = user_data[prot]["AcquisitionType"]
                test_rows = user_data[prot]["Rows"]
                if test_acq_type in ["QSM", "T1w"]:
                    ok, msg = validate_qsm_rows(test_rows) if test_acq_type == "QSM" else validate_t1w_rows(test_rows)
                    if not ok:
                        invalid_protocol = (prot, msg)
                        break

            if invalid_protocol:
                # Show error message for the protocol that failed
                error_line = TABLE_ROWS + len(rows) + 2
                stdscr.addstr(error_line, 0, f"ERROR: Protocol '{invalid_protocol[0]}' is not valid: {invalid_protocol[1]}")
                stdscr.refresh()
                continue  # Return to the loop so the user can fix the issue

            # Count how many are set to skip
            skip_count = sum(1 for p in protocols if user_data[p]["AcquisitionType"] == "Skip")
            if skip_count > 0:
                # Ask user for confirmation
                prompt_line = TABLE_ROWS + len(rows) + 3
                skip_msg = f"{skip_count} protocols are marked skip. Continue anyway? (y/n)"
                stdscr.addstr(prompt_line, 0, skip_msg)
                stdscr.refresh()

                while True:
                    choice = stdscr.getch()
                    if choice in [ord('y'), ord('Y')]:
                        return user_data
                    elif choice in [ord('n'), ord('N')]:
                        break
            else:
                return user_data

        if key == curses.KEY_UP:
            if detail_idx > PROTOCOL_NAME_LINE:
                detail_idx -= 1
            if detail_idx == TABLE_HEADER:
                detail_idx = ACQ_TYPE_LINE

        elif key == curses.KEY_DOWN:
            last_table_line = TABLE_HEADER + len(rows)
            if detail_idx < last_table_line:
                if detail_idx == ACQ_TYPE_LINE and acq_type == "Skip":
                    continue
                detail_idx += 1
            if detail_idx < TABLE_ROWS and detail_idx > ACQ_TYPE_LINE:
                detail_idx = TABLE_ROWS

        elif key == curses.KEY_LEFT or key == curses.KEY_RIGHT:
            dx = 1 if key == curses.KEY_RIGHT else -1

            # If on the protocol line
            if detail_idx == PROTOCOL_NAME_LINE:
                if not valid:
                    continue
                if dx == -1:
                    protocol_idx = max(0, protocol_idx - 1)
                else:
                    protocol_idx = min(len(protocols) - 1, protocol_idx + 1)
                detail_idx = PROTOCOL_NAME_LINE

            elif detail_idx == ACQ_TYPE_LINE:
                possible = ["QSM", "T1w", "Skip"]
                i_acq = possible.index(acq_type)
                new_acq = possible[(i_acq + dx) % len(possible)]
                user_data[protocol]["AcquisitionType"] = new_acq
                acq_type = new_acq
                if new_acq == "QSM":
                    find_and_autoassign_qsm_pairs(rows)
                if new_acq == "T1w":
                    find_and_autoassign_t1w(rows)

            # If we're in the table rows
            elif detail_idx >= TABLE_ROWS:
                row_index = detail_idx - TABLE_ROWS
                if acq_type == "QSM":
                    opts = ["Mag", "Phase", "Real", "Imag", "Skip"]
                elif acq_type == "T1w":
                    opts = ["T1w", "Skip"]
                else:
                    opts = []
                if opts:
                    cur_val = rows[row_index]["Type"]
                    i_val = opts.index(cur_val) if cur_val in opts else 0
                    new_v = (i_val + dx) % len(opts)
                    rows[row_index]["Type"] = opts[new_v]

        if detail_idx < PROTOCOL_NAME_LINE:
            detail_idx = PROTOCOL_NAME_LINE

        last_table_line = TABLE_ROWS + len(rows)
        if detail_idx > last_table_line:
            detail_idx = last_table_line

def interactive_acquisition_selection(grouped):
    def curses_ui(stdscr):
        curses.curs_set(0)

        protocols = grouped["ProtocolName"].unique()
        user_data = {prot: {"AcquisitionType": "Skip", "Rows": []} for prot in protocols}

        protocol_rows_map = {}
        for prot in protocols:
            subset = grouped[grouped["ProtocolName"] == prot]
            row_dicts = []
            for _, row in subset.iterrows():
                row_dicts.append({
                    "SeriesDescription": row["SeriesDescription"],
                    "ImageType": row["ImageType"],
                    "Count": row["Count"],
                    "NumEchoes": row["NumEchoes"],
                    "Type": "Skip"
                })
            protocol_rows_map[prot] = row_dicts

        for prot in protocols:
            user_data[prot]["Rows"] = protocol_rows_map[prot]

        detail_screen(
            stdscr, protocols, user_data,
        )

        return user_data

    return curses.wrapper(curses_ui)

def convert_and_organize(dicom_session, output_dir, dcm2niix_path="dcm2niix"):
    logger = make_logger()
    to_convert = dicom_session[
        (dicom_session['AcquisitionType'] != 'Skip') &
        (dicom_session['Type'] != 'Skip')
    ]

    if to_convert.empty:
        logger.log(LogLevel.ERROR.value, "No acquisitions selected for conversion. Exiting.")
        script_exit(1)
    
    group_cols = [
        "PatientID", "StudyDate", "ProtocolName", "SeriesDescription", 
        "AcquisitionType", "Type", "EchoNumber", "RunNumber"
    ]
    missing_cols = [col for col in group_cols if col not in to_convert.columns]
    if missing_cols:
        logger.log(LogLevel.WARNING.value, f"The following fields are missing from the DICOM session: {', '.join(missing_cols)}")
    if "SeriesInstanceUID" in to_convert.columns:
        group_cols.append("SeriesInstanceUID")
    
    grouped = to_convert.groupby(group_cols)

    if not grouped.groups:
        logger.log(LogLevel.ERROR.value, "No valid acquisitions found. Exiting.")
        script_exit(1)
    
    for grp_keys, grp_data in grouped:
        logger.log(LogLevel.INFO.value, f"Converting {grp_keys}")
        dicom_files = grp_data["DICOM_Path"].unique().tolist()

        tmp_outdir = os.path.join(output_dir, "temp_convert")
        os.makedirs(tmp_outdir, exist_ok=True)
        dcm2niix_base = "temp_output"

        for dicom_file in dicom_files:
            shutil.copy(dicom_file, tmp_outdir)
        
        cmd = f'"{dcm2niix_path}" -o "{tmp_outdir}" -f "{dcm2niix_base}" -z n "{tmp_outdir}"'
        logger.log(LogLevel.INFO.value, f"Running command: '{cmd}'")
        os.system(cmd)
        
        logger.log(LogLevel.INFO.value, "Moving converted NIfTI files into BIDS structure")
        converted_niftis = []
        for fn in os.listdir(tmp_outdir):
            if fn.startswith(dcm2niix_base) and (fn.endswith(".nii") or fn.endswith(".nii.gz")):
                converted_niftis.append(fn)

        for nii_name in converted_niftis:
            row = grp_data.iloc[0]
            subject_id = f"sub-{clean(row['PatientID'])}"
            session_id = f"ses-{clean(row['StudyDate'])}"
            acq_label = f"acq-{clean(row['ProtocolName'])}"
            
            base_name = f"{subject_id}_{session_id}_{acq_label}"
            if row["NumEchoes"] > 1 and row["AcquisitionType"] == "QSM":
                base_name += f"_echo-{int(row['EchoNumber']):02}"
            if row["NumRuns"] > 1:
                base_name += f"_run-{int(row['RunNumber']):02}"
            
            if row["AcquisitionType"] == "QSM":
                base_name += f"_part-{row['Type'].lower()}"
            if row["AcquisitionType"] == "T1w":
                base_name += "_T1w"
            elif row["AcquisitionType"] == "QSM" and row["NumEchoes"] > 1:
                base_name += "_MEGRE"
            elif row["AcquisitionType"] == "QSM" and row["NumEchoes"] == 1:
                base_name += "_T2starw"
            
            ext = ".nii" if nii_name.endswith(".nii") else ".nii.gz"
            out_dir = os.path.join(output_dir, subject_id, session_id, "anat")
            os.makedirs(out_dir, exist_ok=True)
            dst_path = os.path.join(out_dir, f"{base_name}{ext}")
            src_path = os.path.join(tmp_outdir, nii_name)
            shutil.move(src_path, dst_path)
            
            json_src = os.path.join(tmp_outdir, nii_name.replace(ext, ".json"))
            if os.path.isfile(json_src):
                json_dst = os.path.join(out_dir, f"{base_name}.json")
                shutil.move(json_src, json_dst)

        shutil.rmtree(tmp_outdir)

def convert_to_bids(input_dir, output_dir, auto_yes, qsm_protocol_patterns, t1w_protocol_patterns):
    logger = make_logger()
    time_start = time.time()
    dicom_session = load_dicom_session(input_dir, parallel_workers=12, show_progress=True)
    time_end = time.time()
    logger.log(LogLevel.INFO.value, f"Loaded DICOM session in {time_end - time_start:.2f} seconds")
    
    dicom_session.reset_index(drop=True, inplace=True)

    # If we have GE private tags
    if '(0043,102F)' in dicom_session.columns:
        private_map = {0: 'M', 1: 'P', 2: 'REAL', 3: 'IMAGINARY'}
        dicom_session['ImageType'].append(dicom_session['(0043,102F)'].map(private_map).fillna(''))

    # Additional columns
    dicom_session['NumEchoes'] = dicom_session.groupby('ProtocolName')['EchoTime'].transform('nunique')
    dicom_session['EchoNumber'] = dicom_session.groupby('ProtocolName')['EchoTime'].rank(method='dense')
    dicom_session['NumRuns'] = (
        dicom_session
        .groupby(['PatientID','StudyDate','ProtocolName','SeriesDescription', 'ImageType'])['SeriesInstanceUID']
        .transform('nunique')
    )
    dicom_session['RunNumber'] = (
        dicom_session
        .groupby(['PatientID','StudyDate','ProtocolName','SeriesDescription', 'ImageType'])['SeriesInstanceUID']
        .rank(method='dense')
    )

    # Ensure 'AcquisitionType' and 'Type' exist
    if 'AcquisitionType' not in dicom_session.columns:
        dicom_session['AcquisitionType'] = 'Skip'
    if 'Type' not in dicom_session.columns:
        dicom_session['Type'] = 'Skip'

    groupby_fields = ["ProtocolName", "SeriesDescription", "ImageType"]

    grouped = (
        dicom_session
        .groupby(groupby_fields)
        .agg(Count=("InstanceNumber", "count"), NumEchoes=("EchoTime", "nunique"))
        .reset_index()
    )

    # 1) If user is interactive
    if not auto_yes and sys.__stdin__.isatty():
        logger.log(LogLevel.INFO.value, "Entering interactive mode for acquisition selection")
        selections = interactive_acquisition_selection(grouped)
        logger.log(LogLevel.INFO.value, f"User selections:\n{selections}")

        # Write back user selections
        for prot, data_dict in selections.items():
            acq_type = data_dict["AcquisitionType"]
            mask_prot = (grouped['ProtocolName'] == prot)
            grouped.loc[mask_prot, 'AcquisitionType'] = acq_type

            if acq_type in ["QSM","T1w"]:
                for row_info in data_dict["Rows"]:
                    rmask = (
                        mask_prot
                        & (grouped['SeriesDescription'] == row_info['SeriesDescription'])
                        & (grouped['ImageType'] == row_info['ImageType'])
                    )
                    grouped.loc[rmask,'Type'] = row_info['Type']

    # 2) Else do auto-assignment
    else:
        # Auto-assign 'AcquisitionType' based on protocol patterns
        for prot in grouped['ProtocolName'].unique():
            if any(fnmatch.fnmatch(prot.lower(), p.lower()) for p in qsm_protocol_patterns):
                grouped.loc[grouped['ProtocolName']==prot,'AcquisitionType'] = 'QSM'
            elif any(fnmatch.fnmatch(prot.lower(), p.lower()) for p in t1w_protocol_patterns):
                grouped.loc[grouped['ProtocolName']==prot,'AcquisitionType'] = 'T1w'
            # else remains 'Skip'

        # Now group by ProtocolName and run your auto-assign pair logic
        for prot, group_df in grouped.groupby('ProtocolName'):
            logger.log(LogLevel.INFO.value, f"Auto-assigning series for protocol {prot}")
            acq_type = group_df['AcquisitionType'].iloc[0]
            if acq_type == 'QSM':
                # Turn sub-DataFrame into python dicts for find_and_autoassign_qsm_pairs (if that expects a list of dict)
                records = group_df.to_dict('records')
                find_and_autoassign_qsm_pairs(records)
                # Write back
                updated = pd.DataFrame(records)
                logger.log(LogLevel.INFO.value, f"Auto-assignment:\n{updated}")
                # Align on index
                for i, idx in enumerate(group_df.index):
                    grouped.at[idx,'Type'] = updated.at[i,'Type']

            elif acq_type == 'T1w':
                records = group_df.to_dict('records')
                find_and_autoassign_t1w(records)
                # Write back
                updated = pd.DataFrame(records)
                logger.log(LogLevel.INFO.value, f"Auto-assignment:\n{updated}")
                for i, idx in enumerate(group_df.index):
                    grouped.at[idx,'Type'] = updated.at[i,'Type']

    # update dicom_session with the new AcquisitionType and Type
    logger.log(LogLevel.INFO.value, "Merging selections into dataframe...")
    cols_to_drop = ["AcquisitionType", "Type"]
    dicom_session.drop(columns=cols_to_drop, errors='ignore', inplace=True)

    # Now merge only the columns we want from grouped:
    merge_cols = groupby_fields + ["AcquisitionType","Type"]
    dicom_session = dicom_session.merge(
        grouped[merge_cols],
        on=groupby_fields,
        how='left'
    )

    # If some rows had no match in `grouped`, fill them with 'Skip'
    dicom_session['AcquisitionType'] = dicom_session['AcquisitionType'].fillna('Skip')
    dicom_session['Type'] = dicom_session['Type'].fillna('Skip')

    # Finally, do your conversion
    convert_and_organize(dicom_session, output_dir)


def fix_ge_data(bids_dir):
    logger = make_logger()

    logger.log(LogLevel.INFO.value, "Checking for complex or polar data requiring fixing...")
    nifti_session = load_nifti_session(bids_dir, show_progress=True)

    group_cols = [col for col in ["sub", "ses", "acq", "run", "echo", "suffix"] if col in nifti_session.columns]
    grouped = nifti_session.groupby(group_cols)

    for grp_keys, grp_data in grouped:
        if not grp_data["part"].isin(["real", "imag"]).all():
            continue
        
        real_nii_path = grp_data[grp_data["part"] == "real"]["NIfTI_Path"].values[0]
        imag_nii_path = grp_data[grp_data["part"] == "imag"]["NIfTI_Path"].values[0]

        logger.log(LogLevel.INFO.value, f"Fixing complex data (real={real_nii_path}; imag={imag_nii_path})")

        fix_ge_complex(
            real_nii_path=real_nii_path,
            imag_nii_path=imag_nii_path,
            delete_originals=True
        )
    
    for grp_keys, grp_data in grouped:
        if grp_data["part"].isin(["mag", "phase"]).all() and "GE" in grp_data["Manufacturer"].values[0].upper():
            mag_nii_path = grp_data[grp_data["part"] == "mag"]["NIfTI_Path"].values[0]
            phase_nii_path = grp_data[grp_data["part"] == "phase"]["NIfTI_Path"].values[0]

            logger.log(LogLevel.INFO.value, f"Fixing GE polar data (mag={mag_nii_path}; phase={phase_nii_path})")

            fix_ge_polar(
                mag_nii_path=mag_nii_path,
                phase_nii_path=phase_nii_path,
                delete_originals=True
            )

def script_exit(exit_code=0):
    logger = make_logger()
    show_warning_summary(logger)
    logger.log(LogLevel.INFO.value, 'Finished')
    exit(exit_code)

def main():
    parser = argparse.ArgumentParser(
        description="QSMxT dicomConvert: Converts DICOM files to NIfTI/BIDS",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        'input_dir',
        help='Sorted DICOM directory generated using dicom_sort.py of the format {subject}/{session}/{series}'
    )

    parser.add_argument(
        'output_dir',
        help='Output BIDS directory.'
    )

    parser.add_argument(
        '--auto_yes',
        action='store_true',
        help='Run non-interactively if desired.'
    )

    parser.add_argument(
        '--qsm_protocol_patterns',
        default=['*t2starw*', '*qsm*', '*aspire*'],
        nargs='*'
    )

    parser.add_argument(
        '--t1w_protocol_patterns',
        default=['*t1w*', '*mp2rage*'],
        nargs='*'
    )

    args = parser.parse_args()

    args.input_dir = os.path.abspath(args.input_dir)
    args.output_dir = os.path.abspath(args.output_dir)

    os.makedirs(args.output_dir, exist_ok=True)

    logger = make_logger(
        logpath=os.path.join(
            args.output_dir, 
            f"log_{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}.txt"
        ),
        printlevel=LogLevel.INFO,
        writelevel=LogLevel.INFO,
        warnlevel=LogLevel.WARNING,
        errorlevel=LogLevel.ERROR
    )

    logger.log(LogLevel.INFO.value, f"Running QSMxT {get_qsmxt_version()}")
    logger.log(LogLevel.INFO.value, f"Command: {str.join(' ', sys.argv)}")
    logger.log(LogLevel.INFO.value, f"Python interpreter: {sys.executable}")

    diff = get_diff()
    if diff:
        logger.log(
            LogLevel.WARNING.value, 
            f"Working directory not clean! Writing diff to {os.path.join(args.output_dir, 'diff.txt')}..."
        )
        with open(os.path.join(args.output_dir, "diff.txt"), "w") as diff_file:
            diff_file.write(diff)

    with open(os.path.join(args.output_dir, "references.txt"), 'w', encoding='utf-8') as f:
        f.write(f"QSMxT: {get_qsmxt_version()}")
        f.write(f"\nRun command: {str.join(' ', sys.argv)}")
        f.write(f"\nPython interpreter: {sys.executable}")
        f.write("\n\n == References ==")
        # qsmxt, dcm2niix, bids
        f.write("\n\n - Stewart AW, Robinson SD, O'Brien K, et al. QSMxT: Robust masking and artifact reduction for quantitative susceptibility mapping. Magnetic Resonance in Medicine. 2022;87(3):1289-1300. doi:10.1002/mrm.29048")
        f.write("\n\n - Stewart AW, Shaw T, Bollmann S. QSMxT: A Complete QSM Processing Framework. GitHub; 2022. https://github.com/QSMxT/QSMxT")
        f.write("\n\n - Li X, Morgan PS, Ashburner J, Smith J, Rorden C. The first step for neuroimaging data analysis: DICOM to NIfTI conversion. J Neurosci Methods. 2016;264:47-56. doi:10.1016/j.jneumeth.2016.03.001")
        f.write("\n\n - Rorden C et al. Rordenlab/Dcm2niix. GitHub; 2022. https://github.com/rordenlab/dcm2niix")
        f.write("\n\n - Gorgolewski KJ, Auer T, Calhoun VD, et al. The brain imaging data structure, a format for organizing and describing outputs of neuroimaging experiments. Sci Data. 2016;3(1):160044. doi:10.1038/sdata.2016.44")
        f.write("\n\n")

    convert_to_bids(
        input_dir=args.input_dir,
        output_dir=args.output_dir,
        auto_yes=args.auto_yes,
        qsm_protocol_patterns=[pattern.lower() for pattern in args.qsm_protocol_patterns],
        t1w_protocol_patterns=[pattern.lower() for pattern in args.t1w_protocol_patterns]
    )

    fix_ge_data(args.output_dir)

    script_exit()

if __name__ == "__main__":
    main()



import os
import sys
import shutil

from nipype.interfaces.base import CommandLine, traits, TraitedSpec, File, CommandLineInputSpec
from nipype.interfaces.base.traits_extension import isdefined
from qsmxt.scripts.qsmxt_functions import extend_fname

class FastSurferInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="--t1 %s",
        position=0
    )
    num_threads = traits.Int(
        mandatory=False,
        argstr="--parallel --threads %d"
    )


class FastSurferOutputSpec(TraitedSpec):
    out_file = File()


class FastSurferInterface(CommandLine):
    input_spec = FastSurferInputSpec
    output_spec = FastSurferOutputSpec
    _cmd = "run_fastsurfer.sh --sd `pwd` --seg_only --sid output --py python3.8"

    def __init__(self, **inputs):
        super(FastSurferInterface, self).__init__(**inputs)

        self.inputs.on_trait_change(self._num_threads_update, 'num_threads')

        if not isdefined(self.inputs.num_threads):
            self.inputs.num_threads = self._num_threads
        else:
            self._num_threads_update()

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outfile_old = os.path.join('output', 'mri', 'aparc.DKTatlas+aseg.deep.mgz')
        outfile_new = extend_fname(self.inputs.in_file, "_dseg", ext="mgz", out_dir=os.getcwd())
        try:
            shutil.copy(outfile_old, outfile_new)
        except FileNotFoundError:
            print("Expected output from FastSurfer missing! It may have been killed due to insufficient memory.", file=sys.stderr)
        outputs['out_file'] = outfile_new
        return outputs

    def _num_threads_update(self):
        self._num_threads = self.inputs.num_threads
        if self.inputs.num_threads == -1:
            cpu_count = min(8, os.environ["NCPUS"] if "NCPUS" in os.environ else str(os.cpu_count()))
            self.inputs.environ.update({ "OMP_NUM_THREADS" : cpu_count })
        else:
            self.inputs.environ.update({ "OMP_NUM_THREADS" : f"{self.inputs.num_threads}" })

   



#!/usr/bin/env pytest
import os
import glob
import tempfile
import pytest
import shutil

import numpy as np
import qsm_forward
from qsmxt.cli.main import main
from qsmxt.scripts.logger import LogLevel, make_logger
from qsmxt.scripts.qsmxt_functions import get_qsm_premades
from qsmxt.scripts.sys_cmd import sys_cmd
from qsmxt.tests.utils import *

run_workflows = True

def getrunid():
    return os.environ.get('RUN_ID') or ''

def gettempdir():
    return os.environ.get('TEST_DIR') or tempfile.gettempdir()

@pytest.fixture
def bids_dir_public():
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== BIDS Preparation ===")

    tmp_dir = gettempdir()
    bids_dir = os.path.join(tmp_dir, "bids")
    if not glob.glob(os.path.join(bids_dir, "sub*1")):
        logger.log(LogLevel.INFO.value, f"No subjects in BIDS directory yet")
        head_phantom_maps_dir = os.path.join(tmp_dir, 'data')
        if not os.path.exists(head_phantom_maps_dir):
            logger.log(LogLevel.INFO.value, f"Head phantom maps directory does not exist yet")
            if not os.path.exists(os.path.join(tmp_dir, 'head-phantom-maps.tar')):
                logger.log(LogLevel.INFO.value, f"Head phantom maps tar file does not exist - downloading...")
                download_from_osf(
                    project="9jc42",
                    local_path=os.path.join(tmp_dir, "head-phantom-maps.tar")
                )
            logger.log(LogLevel.INFO.value, f"Extracting then deleting head-phantom-maps.tar...")
            sys_cmd(f"tar xf {os.path.join(tmp_dir, 'head-phantom-maps.tar')} -C {tmp_dir}")

        logger.log(LogLevel.INFO.value, "Preparing simulation information...")
        tissue_params = qsm_forward.TissueParams(os.path.join(tmp_dir, 'data'))
        recon_params_all = [
            qsm_forward.ReconParams(voxel_size=np.array([1.0, 1.0, 1.0]), session=session, TEs=TEs, TR=TR, flip_angle=flip_angle, suffix=suffix, save_phase=save_phase)
            for (session, TEs, TR, flip_angle, suffix, save_phase) in [
                ("1", np.array([3.5e-3]), 7.5e-3, 40, "T1w", False),
                ("1", np.array([0.012, 0.020]), 0.05, 15, "T2starw", True),
                ("2", np.array([0.012, 0.020]), 0.05, 15, "T2starw", True)
            ]
        ]

        logger.log(LogLevel.INFO.value, "Generating BIDS dataset...")
        for recon_params in recon_params_all:
            qsm_forward.generate_bids(tissue_params=tissue_params, recon_params=recon_params, bids_dir=bids_dir)

    return bids_dir

@pytest.fixture
def bids_dir_real():
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== BIDS Preparation ===")
    tmp_dir = gettempdir()
    if not os.path.exists(os.path.join(tmp_dir, 'bids-secret')):
        logger.log(LogLevel.INFO.value, f"Secret BIDS directory doesn not exist!")
        if not os.path.exists(os.path.join(tmp_dir, 'bids-secret.zip')):
            logger.log(LogLevel.INFO.value, f"bids-secret.zip does not exist! Downloading...")
            download_from_osf(
                project="n6uqk",
                local_path=os.path.join(tmp_dir, "bids-secret.zip")
            )
        
        logger.log(LogLevel.INFO.value, f"Extracting bids-secret.zip...")
        sys_cmd(f"unzip {os.path.join(tmp_dir, 'bids-secret.zip')} -d {tmp_dir}")
        sys_cmd(f"rm {os.path.join(tmp_dir, 'bids-secret.zip')}")

    return os.path.join(tmp_dir, 'bids-secret')

@pytest.mark.parametrize("premade", [
    p for p in get_qsm_premades().keys() if p != 'default'
])
def test_premade(bids_dir_public, premade):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING PREMADE {premade} ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    args = [
        bids_dir,
        "--premade", premade,
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run if necessary
    args = main(args)
    
    # upload output folder
    tar_file = compress_folder(folder=args.output_dir, result_id=premade)
    sys_cmd(f"rm -rf {os.path.join(gettempdir(), 'public-outputs')}")
    os.makedirs(os.path.join(gettempdir(), "public-outputs"), exist_ok=True)
    shutil.move(tar_file, os.path.join(gettempdir(), "public-outputs", tar_file))

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"Premade {premade}")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_Chimap.nii*')[0], title=premade, colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)


def test_nocombine(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING NO PHASE COMBINATION ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    args = [
        bids_dir,
        "--premade", "fast",
        "--combine_phase", "off",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run if necessary
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_nocombine")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_Chimap.nii*')[0], title='No combine', colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)


def test_nomagnitude(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING NO MAGNITUDE ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)
    for mag_file in glob.glob(os.path.join(bids_dir, "sub-1", "ses-1", "anat", "*mag*")):
        os.remove(mag_file)
    
    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--premade", "fast",
        "--masking_input", "magnitude",
        "--num_echoes", "1",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run - it should fall back to phase-based masking with a warning
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_nomagnitude")
        qsm_result = find_files(os.path.join(args.output_dir), '*_Chimap.nii*')[0]
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=qsm_result, title='No magnitude', colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_inhomogeneity_correction(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING INHOMOGENEITY CORRECTION ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--premade", "fast",
        "--masking_algorithm", "threshold",
        "--filling_algorithm", "bet",
        "--inhomogeneity_correction",
        "--num_echoes", "1",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_inhomogeneity_correction")
        qsm_result = find_files(args.output_dir, '*_Chimap.nii*')[0]
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=qsm_result, title='Inhomogeneity correction', colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_hardcoded_percentile_threshold(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING HARDCODED PERCENTILE THRESHOLD ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--premade", "fast",
        "--masking_algorithm", "threshold",
        "--masking_input", "magnitude",
        "--threshold_value", "0.25",
        "--num_echoes", "1",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_hardcoded_percentile_threshold")
        qsm_result = find_files(args.output_dir, '*_Chimap.nii*')[0]
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=qsm_result, title='Hardcoded percentile threshold (0.25)', colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_hardcoded_absolute_threshold(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING HARDCODED ABSOLUTE THRESHOLD ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--premade", "fast",
        "--masking_algorithm", "threshold",
        "--masking_input", "magnitude",
        "--threshold_value", "15",
        "--num_echoes", "1",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_hardcoded_absolute_threshold")
        qsm_result = find_files(args.output_dir, '*_Chimap.nii*')[0]
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=qsm_result, title='Hardcoded absolute threshold (15)', colorbar=True, vmin=-0.1, vmax=+0.1, out_png='qsm.png', cmap='gray'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_use_existing_masks(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING EXISTING MASKS ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)
    
    args = [
        bids_dir,
        "--use_existing_masks",
        "--premade", "fast",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    args = main(args)

    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_use_existing_masks")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)


def test_supplementary_images(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING SUPPLEMENTARY IMAGES AND DICOMS ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)
    
    args = [
        bids_dir,
        "--do_qsm",
        "--do_swi",
        "--do_t2starmap",
        "--do_r2starmap",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1",
        "--export_dicoms"
    ]
    
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_supplementary_images")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_swi.nii*')[0], title='SWI', out_png='swi.png', cmap='gray'))})")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_minIP.nii*')[0], title='SWI MIP', out_png='swi_mip.png', cmap='gray'))})")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_T2starmap.nii*')[0], title='T2* map', out_png='t2s.png', cmap='gray', vmin=0, vmax=100, colorbar=True, cbar_label='T2* (ms)'))})")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_R2starmap.nii*')[0], title='R2* map', out_png='r2s.png', cmap='gray', vmin=0, vmax=100, colorbar=True, cbar_label='R2* (ms)'))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)
    

def test_realdata(bids_dir_real):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING REAL DATA ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids-real")
    shutil.copytree(bids_dir_real, bids_dir)

    args = [
        bids_dir,
        "--premade", "fast",
        "--auto_yes",
        "--debug"
    ]
    
    args = main(args)
    local_path = compress_folder(folder=args.output_dir, result_id='real')

    try:
        upload_to_rdm(
            local_path=local_path,
            remote_path=f"QSMFUNCTOR-Q0748/data/QSMxT-Test-Results/{os.path.split(local_path)[1]}"
        )
        github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
        if github_step_summary:
            write_to_file(github_step_summary, f"test_realdata")
            write_to_file(github_step_summary, "Results uploaded to RDM")
            for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
                write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")
    except:
        github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
        if not github_step_summary:
            logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
        else:
            write_to_file(github_step_summary, f"test_realdata")
            write_to_file(github_step_summary, "Result upload to RDM failed!")
            for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
                write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_singleecho(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING SINGLE ECHO WITH PHASE COMBINATION / ROMEO ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--unwrapping_algorithm", "romeo",
        "--num_echoes", "1",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_singleecho")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_Chimap.nii*')[0], title='Single echo', out_png='qsm.png', cmap='gray', colorbar=True, vmin=-0.1, vmax=+0.1))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)

def test_laplacian_and_tv(bids_dir_public):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"=== TESTING LAPLACIAN UNWRAPPING AND TV ALGO ===")

    bids_dir = os.path.join(gettempdir(), f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{getrunid()}-bids")
    shutil.copytree(bids_dir_public, bids_dir)

    # run pipeline and specifically choose magnitude-based masking
    args = [
        bids_dir,
        "--unwrapping_algorithm", "laplacian",
        "--qsm_algorithm", "tv",
        "--combine_phase", "off",
        "--auto_yes",
        "--debug",
        "--subjects", "sub-1",
        "--sessions", "ses-1"
    ]
    
    # create the workflow and run
    args = main(args)

    # generate image
    github_step_summary = os.environ.get('GITHUB_STEP_SUMMARY')
    if not github_step_summary:
        logger.log(LogLevel.WARNING.value, f"GITHUB_STEP_SUMMARY variable not found! Cannot write summary.")
    else:
        write_to_file(github_step_summary, f"test_laplacian_and_tv")
        write_to_file(github_step_summary, f"![result]({upload_png(display_nii(nii_path=find_files(args.output_dir, '*_Chimap.nii*')[0], title='Laplacian + TV', out_png='qsm.png', cmap='gray', colorbar=True, vmin=-0.1, vmax=+0.1))})")
        for png_file in glob.glob(os.path.join(args.output_dir, '*.png')):
            write_to_file(github_step_summary, f"![summary]({upload_png(png_file)})")

    shutil.rmtree(args.bids_dir)





#!/usr/bin/env python3

# Adapted from Alex Weston
# Digital Innovation Lab, Mayo Clinic
# https://gist.github.com/alex-weston-13/4dae048b423f1b4cb9828734a4ec8b83

import argparse
import os
import sys
import pydicom  # pydicom is using the gdcm package for decompression
import datetime
import re
import shutil

from qsmxt.scripts.qsmxt_functions import get_qsmxt_version, get_diff
from qsmxt.scripts.logger import LogLevel, make_logger, show_warning_summary

def empty_dirs(root_dir='.', recursive=True):
    empty_dirs = []
    for root, dirs, files in os.walk(root_dir, topdown=False):
        #print root, dirs, files
        if recursive:
            all_subs_empty = True  # until proven otherwise
            for sub in dirs:
                full_sub = os.path.join(root, sub)
                if full_sub not in empty_dirs:
                    #print full_sub, "not empty"
                    all_subs_empty = False
                    break
        else:
            all_subs_empty = (len(dirs) == 0)
        if all_subs_empty and len(files) == 0:
            empty_dirs.append(root)
            yield root

def find_empty_dirs(root_dir='.', recursive=True):
    return list(empty_dirs(root_dir, recursive))

def clean_text(string):
    return re.sub(r'[^a-zA-Z0-9_-]', '_', string)

def find_dicoms(input_dir, check_all_files):
    logger = make_logger()
    unsortedList = []
    for root, dirs, files in os.walk(input_dir):
        for f in files:
            file_extension = ".".join(f.split('.')[-1:])
            if file_extension.lower() in ['ima', 'dcm']:
                unsortedList.append(os.path.join(root, f))
    if not unsortedList or check_all_files:
        if not unsortedList:
            logger.log(LogLevel.WARNING.value, "No .IMA or .dcm files found! Checking all files for valid DICOM headers...")
        for root, dirs, files in os.walk(input_dir):
            for f in files:
                try:
                    ds = pydicom.read_file(os.path.join(root, f), force=True)
                    if len(ds) > 1:
                        unsortedList.append(os.path.join(root, f))
                except:
                    pass
    return unsortedList

def dicomsort(input_dir, output_dir, use_patient_names, use_session_incrementer, check_all_files, delete_originals):
    logger = make_logger()
    os.makedirs(output_dir, exist_ok=True)
    logger.log(LogLevel.INFO.value, "Reading file list...")
    unsortedList = find_dicoms(input_dir, check_all_files)
    logger.log(LogLevel.INFO.value, f"{len(unsortedList)} DICOM files found.")
    fail = False

    subjName_dates = []
    subjName_sessionNums = {}

    logger.log(LogLevel.INFO.value, f"Sorting DICOMs in {output_dir}...")
    for dicom_loc in unsortedList:
        # read the file
        try:
            ds = pydicom.read_file(dicom_loc, force=True)
        except:
            logger.log(LogLevel.WARNING.value, f"Failed to read file as DICOM: {dicom_loc}. Skipping...")
            continue
    
        # get patient, study, and series information
        patientName = clean_text(str(ds.get("PatientName", "NA")))
        patientID = clean_text(ds.get("PatientID", "NA"))
        studyDate = clean_text(ds.get("StudyDate", "NA"))
        studyDescription = clean_text(ds.get("StudyDescription", "NA"))
        protocolName = clean_text(ds.get("ProtocolName", "NA"))
        seriesNumber = clean_text(str(ds.get("SeriesNumber", "NA")))
        
        # generate new, standardized file name
        modality = ds.get("Modality","NA")
        studyInstanceUID = ds.get("StudyInstanceUID","NA")
        seriesInstanceUID = ds.get("SeriesInstanceUID","NA")
        instanceNumber = str(ds.get("InstanceNumber","0"))
        fileName = modality + "." + seriesInstanceUID + "." + instanceNumber + ".dcm"

        subj_name = patientName if use_patient_names else patientID
        subj_name = subj_name.replace('-', '').replace('_', '')
        subj_name = subj_name.replace('-', '').replace('_', '')
        if not subj_name:
            subj_name = "NA"
        
        # save files to a 3-tier nested folder structure
        subjFolderName = f"sub-{subj_name}"
        seriesFolderName = f"series-{seriesNumber}_{protocolName}"
        subjName_date = f"{subj_name}_{studyDate}"

        if not any(subj_name in x for x in subjName_dates):
            logger.log(LogLevel.INFO.value, f"Identified subject: {subj_name}")
        
        if subjName_date not in subjName_dates:
            subjName_dates.append(subjName_date)
            if subj_name in subjName_sessionNums.keys():
                subjName_sessionNums[subj_name] += 1
            else:
                subjName_sessionNums[subj_name] = 1
            logger.log(LogLevel.INFO.value, f"Identified session: {subj_name} #{subjName_sessionNums[subj_name]} {studyDate}")

        sesFolderName = f"ses-{subjName_sessionNums[subj_name]}" if (use_session_incrementer or studyDate == "") else f"ses-{studyDate}"
        
        if not os.path.exists(os.path.join(output_dir, subjFolderName, sesFolderName, seriesFolderName)):
            logger.log(LogLevel.INFO.value, f"Identified series: {subjFolderName}/{sesFolderName}/{seriesFolderName}")
            os.makedirs(os.path.join(output_dir, subjFolderName, sesFolderName, seriesFolderName), exist_ok=True)

        # copy the file to the new location, retaining all metadata
        shutil.copy2(dicom_loc, os.path.join(output_dir, subjFolderName, sesFolderName, seriesFolderName, fileName))

        if not os.path.exists(os.path.join(output_dir, subjFolderName, sesFolderName, seriesFolderName, fileName)):
            fail = True

    if not fail and delete_originals:
        for dicom_loc in unsortedList:
            os.remove(dicom_loc)


def main():
    parser = argparse.ArgumentParser(
        description="QSMxT dicomSort: Sorts DICOM files into a folder structure of the form <out_dir>/<Subject>/<Session>/<Series>",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        'input_dir',
        help='Input DICOM directory to be recursively searched for DICOM files.'
    )

    parser.add_argument(
        'output_dir',
        help='Output directory to house sorted DICOMs.'
    )

    parser.add_argument(
        '--use_patient_names',
        action='store_true',
        help='Use the DICOM \'PatientName\' field rather than \'PatientID\' to identify subjects.'
    )

    parser.add_argument(
        '--use_session_incrementer',
        action='store_true',
        help='Use an incrementer to identify scanning sessions rather than the \'SessionDate\' DICOM field.'
    )

    parser.add_argument(
        '--check_all_files',
        action='store_true',
        help='Ignores the DICOM file extensions .dcm and .IMA and instead reads all files for valid DICOM headers. '+
             'This is useful if some of your DICOM files have unusual file extensions or none at all.'
    )

    parser.add_argument(
        '--delete_originals',
        action='store_true',
        help='Delete the original DICOM files and folders after successfully sorting; by ' +
             'default this is on when input_dir == output_dir'
    )

    args = parser.parse_args()

    args.input_dir = os.path.abspath(args.input_dir)
    args.output_dir = os.path.abspath(args.output_dir)

    os.makedirs(args.output_dir, exist_ok=True)

    logger = make_logger(
        logpath=os.path.join(args.output_dir, f"log_{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}.txt"),
        printlevel=LogLevel.INFO,
        writelevel=LogLevel.INFO,
        warnlevel=LogLevel.WARNING,
        errorlevel=LogLevel.ERROR
    )

    logger.log(LogLevel.INFO.value, f"Running QSMxT {get_qsmxt_version()}")
    logger.log(LogLevel.INFO.value, f"Command: {str.join(' ', sys.argv)}")
    logger.log(LogLevel.INFO.value, f"Python interpreter: {sys.executable}")

    diff = get_diff()
    if diff:
        logger.log(LogLevel.WARNING.value, f"Working directory not clean! Writing diff to {os.path.join(args.output_dir, 'diff.txt')}...")
        with open(os.path.join(args.output_dir, "diff.txt"), "w") as diff_file:
            diff_file.write(diff)

    with open(os.path.join(args.output_dir, "references.txt"), 'w', encoding='utf-8') as f:
        # output QSMxT version, run command, and python interpreter
        f.write(f"QSMxT: {get_qsmxt_version()}")
        f.write(f"\nRun command: {str.join(' ', sys.argv)}")
        f.write(f"\nPython interpreter: {sys.executable}")

        f.write("\n\n == References ==")

        # qsmxt, pydicom, sort_dicoms.py
        f.write("\n\n - Stewart AW, Robinson SD, O'Brien K, et al. QSMxT: Robust masking and artifact reduction for quantitative susceptibility mapping. Magnetic Resonance in Medicine. 2022;87(3):1289-1300. doi:10.1002/mrm.29048")
        f.write("\n\n - Stewart AW, Bollman S, et al. QSMxT/QSMxT. GitHub; 2022. https://github.com/QSMxT/QSMxT")
        f.write("\n\n - Mason D, scaramallion, mrbean-bremen, et al. Pydicom/Pydicom: Pydicom 2.3.0. Zenodo; 2022. doi:10.5281/zenodo.6394735")
        f.write("\n\n - Weston A. alex-weston-13/sort_dicoms.py. GitHub; 2020. https://gist.github.com/alex-weston-13/4dae048b423f1b4cb9828734a4ec8b83")
        f.write("\n\n")

    dicomsort(
        input_dir=args.input_dir,
        output_dir=args.output_dir,
        use_patient_names=args.use_patient_names,
        use_session_incrementer=args.use_session_incrementer,
        check_all_files=args.check_all_files,
        delete_originals=args.input_dir == args.output_dir or args.delete_originals
    )

    show_warning_summary(logger)

    logger.log(LogLevel.INFO.value, 'Finished')
    



import os
import logging as _logging
from enum import Enum

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

class _StringStream:
    '''
    a stream used to keep track of messages in a list
    '''

    def __init__(self, max_records=None, print_new_records=True):
        self.items = []
        self.max_records = max_records
        self.print_new_records = print_new_records

    def __get__(self, index):
        return self.items[index]

    def __len__(self):
        return len(self.items)

    def write(self, record):
        self.items.append(record)

        if self.print_new_records:
            print(record, end="")

        if self.max_records and len(self.items) > self.max_records:
            self.items.pop(0)

    def flush(self):
        pass

def make_logger(name='main', logpath=None, printlevel=LogLevel.INFO, warnlevel=LogLevel.WARNING, errorlevel=LogLevel.ERROR, writelevel=LogLevel.INFO):
    # create/get logger
    logger = _logging.getLogger(name=name)

    # check level names if needed
    for log_level in LogLevel:
        if log_level.value not in _logging._levelToName.values():
            _logging.addLevelName(log_level.value, log_level.name)

    # return logger if it already has 3 handlers and we aren't writing to file
    if logger.hasHandlers() and (len(logger.handlers) == 3 and logpath is None):
        return logger

    # return logger with updated baseFilename path if necessary if there are 4 handlers
    if logpath: logpath = os.path.abspath(logpath)
    if logger.hasHandlers() and len(logger.handlers) == 4:
        logger.handlers[3].baseFilename == logpath
        return logger

    # create handlers
    console_handler = _logging.StreamHandler(stream=_StringStream())
    warnings_handler = _logging.StreamHandler(stream=_StringStream(print_new_records=False))
    errors_handler = _logging.StreamHandler(stream=_StringStream(print_new_records=False))
    if logpath:
        file_handler = _logging.FileHandler(logpath, mode='w')

    # create formatter
    # formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # https://docs.python.org/3.7/library/logging.html#logrecord-attributes
    formatter = _logging.Formatter('[%(levelname)s]: %(message)s')

    # add formatters to handlers
    console_handler.setFormatter(formatter)
    warnings_handler.setFormatter(formatter)
    errors_handler.setFormatter(formatter)
    if logpath:
        file_handler.setFormatter(formatter)

    # add handlers to logger
    if logger.hasHandlers() == False or len(logger.handlers) == 0:
        logger.addHandler(console_handler)
        logger.addHandler(warnings_handler)
        logger.addHandler(errors_handler)
    if logpath:
        logger.addHandler(file_handler)

    # set log levels
    logger.handlers[0].setLevel(printlevel.value)
    logger.handlers[1].setLevel(warnlevel.value)
    logger.handlers[2].setLevel(errorlevel.value)
    if logpath:
        logger.handlers[3].setLevel(writelevel.value)
    logger.setLevel(printlevel.value)

    return logger


def show_log(logger):
    for message in logger.handlers[0].stream.items:
        print(message, end='')

def show_warning_summary(logger):
    warnings_occurred = False
    errors_occurred = False
    
    if len(logger.handlers) > 1:
        for message in logger.handlers[1].stream.items:
            if "WARNING" in message:
                warnings_occurred = True
                break
    
    if len(logger.handlers) > 2:
        for message in logger.handlers[2].stream.items:
            if "ERROR" in message:
                errors_occurred = True
                break
                
    if warnings_occurred:
        logger.log(LogLevel.INFO.value, "Warnings occurred!")
        
    if errors_occurred:
        logger.log(LogLevel.INFO.value, "Errors occurred!")




from nipype.interfaces.base import TraitedSpec, File, traits, InputMultiPath
from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from qsmxt.scripts import qsmxt_functions
import os

class PbMaskingInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(PbMaskingInputSpec, self).__init__(**inputs)
    phase = File(
        exists=True,
        mandatory=True,
        argstr="--phase %s"
    )
    mask = File(
        argstr="--output %s",
        name_source=['phase'],
        name_template='%s_pb_mask.nii'
    )

class PbMaskingOutputSpec(TraitedSpec):
    mask = File()

class PbMaskingInterface(CommandLineJulia):
    def __init__(self, **inputs): super(PbMaskingInterface, self).__init__(**inputs)
    input_spec = PbMaskingInputSpec
    output_spec = PbMaskingOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "hagberg_pb_masking.jl")


class RomeoMaskingInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(RomeoMaskingInputSpec, self).__init__(**inputs)
    phase = InputMultiPath(
        exists=True,
        mandatory=True,
        argstr="--phase %s"
    )
    magnitude = InputMultiPath(
        mandatory=False,
        exists=True,
        argstr="--mag %s"
    )
    TEs = traits.ListFloat(
        mandatory=False,
        argstr="--TEs '[%s]'"
    )
    TE = traits.Float(
        mandatory=False,
        argstr="--TEs '[%s]'"
    )
    weight_type = traits.Str(
        default_value="grad+second",
        argstr="--type %s"
    )
    quality_map = File(
        argstr="--output %s",
        name_source=['phase'],
        name_template='%s_romeo_voxelquality.nii'
    )

class RomeoMaskingOutputSpec(TraitedSpec):
    quality_map = File()

class RomeoMaskingInterface(CommandLineJulia):
    def __init__(self, **inputs): super(RomeoMaskingInterface, self).__init__(**inputs)
    input_spec = RomeoMaskingInputSpec
    output_spec = RomeoMaskingOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "romeo_voxelquality.jl")

    def _format_arg(self, name, trait_spec, value):
        if name == 'TEs' or name == 'TE':
            if self.inputs.TEs is None and self.inputs.TE is None:
                raise ValueError("Either TEs or TE must be provided")
        return super(RomeoMaskingInterface, self)._format_arg(name, trait_spec, value)
    



#!/usr/bin/env python3

import json
import os
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, traits

class JsonInputSpec(BaseInterfaceInputSpec):
    in_dict = traits.Dict(mandatory=True)
    out_file = File()

class JsonOutputSpec(TraitedSpec):
    out_file = File()

class JsonInterface(SimpleInterface):
    input_spec = JsonInputSpec
    output_spec = JsonOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = os.path.abspath(self.inputs.out_file)
        with open(os.path.abspath(self.inputs.out_file), 'w', encoding='utf-8') as json_file:
            json.dump(self.inputs.in_dict, json_file)
        return runtime




from nipype.interfaces.base import CommandLine, TraitedSpec, traits, File, CommandLineInputSpec


class Bet2InputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0
    )
    out_file = File(
        argstr="%s",
        name_source=['in_file'],
        name_template='%s_bet.nii.gz',
        position=1,
        exists=False
    )
    mask = File(
        argstr="-m %s",
        name_source=['in_file'],
        name_template='%s_bet-mask.nii.gz',
        position=2,
        exists=False
    )
    fractional_intensity = traits.Float(
        mandatory=False,
        argstr="-f %f",
        default=0.5
    )


class Bet2OutputSpec(TraitedSpec):
    out_file = File(exists=True)
    mask = File(exists=True)


class Bet2Interface(CommandLine):
    input_spec = Bet2InputSpec
    output_spec = Bet2OutputSpec
    _cmd = "bet"




#!/usr/bin/env python3
import nibabel as nib
import numpy as np
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File


def mgz2nii(in_file, out_file=None):
    if not out_file: out_file = in_file.replace(".mgz", "_nii.nii")
    mgz = nib.load(in_file)

    # for some reason nibabel needs help converting the data type properly to numpy
    dtypes = { # https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
        0 : np.uint8, # UCHAR
        1 : int,      # INT
        3 : float,    # FLOAT
        4 : np.short  # SHORT
    }

    data = np.array(mgz.get_fdata(), dtype=dtypes[int(mgz.header['type'])])
    
    nii = nib.Nifti1Image(data, affine=mgz.affine, header=mgz.header)
    nib.save(nii, out_file)

    return out_file

class Mgz2NiiInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True)


class Mgz2NiiOutputSpec(TraitedSpec):
    out_file = File()


class Mgz2NiiInterface(SimpleInterface):
    input_spec = Mgz2NiiInputSpec
    output_spec = Mgz2NiiOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = mgz2nii(self.inputs.in_file)
        return runtime


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'in_file',
        type=str
    )

    parser.add_argument(
        'out_file',
        nargs='?',
        default=None,
        const=None,
        type=str
    )

    args = parser.parse_args()
    out_file = mgz2nii(args.in_file, args.out_file)
